#!/usr/bin/env python
"""
SaveNormalizedTab class
"""

from loguru import logger

from ibeatles import DataType
from ibeatles.session import SessionKeys, SessionSubKeys
from ibeatles.utilities.pyqrgraph import Pyqtgrah as PyqtgraphUtilities

from .save_tab import SaveTab


class SaveNormalizedTab(SaveTab):
    def normalized(self):
        """record the ROI selected"""

        data_type = DataType.normalized

        list_files = self.parent.list_files[data_type]
        current_folder = self.parent.data_metadata[data_type]["folder"]
        time_spectra_filename = self.parent.data_metadata[data_type]["time_spectra"]["filename"]
        list_files_selected = [int(index) for index in self.parent.list_file_selected[data_type]]
        list_roi = self.parent.list_roi[data_type]

        o_pyqt = PyqtgraphUtilities(
            parent=self.parent,
            image_view=self.parent.ui.normalized_image_view,
            data_type=data_type,
        )
        state = o_pyqt.get_state()
        o_pyqt.save_histogram_level()
        histogram = self.parent.image_view_settings[data_type]["histogram"]

        logger.info("Recording parameters of normalized tab")
        logger.info(f" len(list files) = {len(list_files)}")
        logger.info(f" current folder: {current_folder}")
        logger.info(f" time spectra filename: {time_spectra_filename}")
        logger.info(f" list files selected: {list_files_selected}")
        logger.info(f" len(list rois): {len(list_roi)}")
        logger.info(f" state: {state}")
        logger.info(f" histogram: {histogram}")

        self.session_dict[data_type][SessionSubKeys.list_files] = list_files
        self.session_dict[data_type][SessionSubKeys.current_folder] = current_folder
        self.session_dict[data_type][SessionSubKeys.time_spectra_filename] = time_spectra_filename
        self.session_dict[data_type][SessionSubKeys.list_files_selected] = list_files_selected
        self.session_dict[data_type][SessionSubKeys.list_rois] = list_roi
        self.session_dict[data_type][SessionSubKeys.image_view_state] = state
        self.session_dict[data_type][SessionSubKeys.image_view_histogram] = histogram
        self.session_dict[SessionKeys.reduction] = self.parent.session_dict[SessionKeys.reduction]



from .. import DataType
from ..utilities.pyqrgraph import Pyqtgrah as PyqtgraphUtilities


class BinningHandler(object):
    def __init__(self, parent=None):
        self.parent = parent
        self.binning_ui = self.parent.binning_ui

    def display_image(self, data=None):
        o_pyqt = PyqtgraphUtilities(
            parent=self.parent,
            image_view=self.parent.binning_line_view["image_view"],
            data_type=DataType.bin,
        )
        _state = o_pyqt.get_state()
        o_pyqt.save_histogram_level(data_type_of_data=DataType.normalized)

        if data is not None:
            self.binning_ui.data = data
            self.parent.binning_line_view["image_view"].setImage(data)
        else:
            data = self.parent.binning_ui.data
            if len(data) == 0:
                return
            else:
                self.parent.binning_line_view["image_view"].setImage(data)
                self.binning_ui.data = data

        o_pyqt.set_state(_state)
        o_pyqt.reload_histogram_level()

    def display_selection(self):
        if self.parent.binning_line_view["ui"]:
            binning_line_view = self.parent.binning_line_view
            line_view_binning = binning_line_view["ui"]
            # self.parent.binning_ui.line_view = line_view_binning

            pos = binning_line_view["pos"]
            adj = binning_line_view["adj"]
            lines = binning_line_view["pen"]

            self.parent.binning_line_view["ui"].setData(pos=pos, adj=adj, pen=lines, symbol=None, pxMode=False)

            self.parent.binning_line_view["ui"] = line_view_binning
            self.parent.binning_line_view["pos"] = pos
            self.parent.binning_line_view["adj"] = adj
            self.parent.binning_line_view["pen"] = lines

            # self.parent.binning_ui.line_view.setData(pos=pos,
            # adj=adj,
            # pen=lines,
            # symbol=None,
            # pxMode=False)

            # self.parent.binning_ui.line_view_binning = line_view_binning
            # self.parent.binning_ui.pos = pos
            # self.parent.binning_ui.adj = adj
            # self.parent.binning_ui.lines = lines



#!/usr/bin/env python
"""
Log bin
"""

import logging

import numpy as np

from ibeatles.tools.utilities import TimeSpectraKeys


class LogBin:
    log_bins = {
        TimeSpectraKeys.tof_array: None,
        TimeSpectraKeys.file_index_array: None,
        TimeSpectraKeys.lambda_array: None,
    }

    def __init__(self, parent=None, source_radio_button=TimeSpectraKeys.file_index_array):
        self.parent = parent
        self.source_array = source_radio_button

    def create_log_file_index_bin_array(self, bin_value=1):
        """
        This method create the 1D array of the bin positions
        :param source_array: either 'file_index', 'lambda' or 'tof'
        :param bin_value: value of the logarithmic bin
        """
        original_array = np.array(self.parent.time_spectra[self.source_array])

        # create the log bin array [value1, value2, value3, value4....]
        start_parameter = original_array[0]
        parameter_end = original_array[-1]
        new_bin_array = [start_parameter]
        parameter = start_parameter
        while parameter <= parameter_end:
            parameter += parameter * bin_value
            new_bin_array.append(parameter)

        # in case the last bin is smaller than the last array value
        if new_bin_array[-1] <= original_array[-1]:
            parameter += parameter * bin_value
            new_bin_array.append(parameter)

        self.parent.full_bin_axis_requested = new_bin_array

        logging.info(f"new bins array: {new_bin_array}")

        # we need to find where the file index end up in this array
        # will create [[0],[],[],[1],[2,3],[4,5,6,7],...]
        index_of_bin = [[] for _ in np.arange(len(new_bin_array) - 1)]
        for _bin_index, _bin in enumerate(original_array):
            result = np.where(_bin >= new_bin_array)
            try:
                index = result[0][-1]
                index_of_bin[index].append(_bin_index)
            except IndexError:
                continue

        array_of_bins = index_of_bin

        self.log_bins[self.source_array] = array_of_bins
        logging.info(f"log {self.source_array} array of bins: {array_of_bins}")

    def create_log_bin_arrays(self):
        logging.info("Creating the other arrays")

        file_index_array_of_bins = self.log_bins[self.source_array]

        original_tof_array = np.array(self.parent.time_spectra[TimeSpectraKeys.tof_array])
        original_lambda_array = np.array(self.parent.time_spectra[TimeSpectraKeys.lambda_array])

        log_bins_tof_array = []
        log_bins_lambda_array = []

        for _index, _bin in enumerate(file_index_array_of_bins):
            if _bin == []:
                log_bins_tof_array.append([])
                log_bins_lambda_array.append([])
                continue

            tof_bin = []
            lambda_bin = []
            for _file_index in _bin:
                tof_bin.append(original_tof_array[_file_index])
                lambda_bin.append(original_lambda_array[_file_index])
            log_bins_tof_array.append(tof_bin)
            log_bins_lambda_array.append(lambda_bin)

        self.log_bins[TimeSpectraKeys.tof_array] = log_bins_tof_array
        self.log_bins[TimeSpectraKeys.lambda_array] = log_bins_lambda_array
        self.log_bins[TimeSpectraKeys.file_index_array] = file_index_array_of_bins

    def get_log_file_index(self):
        return self.log_bins[TimeSpectraKeys.file_index_array]

    def get_log_tof(self):
        return self.log_bins[TimeSpectraKeys.tof_array]

    def get_log_lambda(self):
        return self.log_bins[TimeSpectraKeys.lambda_array]



#!/usr/bin/env python
"""
Time spectra module
"""

import glob
import os
from pathlib import Path

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
from neutronbraggedge.experiment_handler.experiment import Experiment
from neutronbraggedge.experiment_handler.tof import TOF
from qtpy.QtWidgets import QMainWindow, QSizePolicy, QVBoxLayout

from ibeatles import load_ui
from ibeatles.tools.utilities import TimeSpectraKeys
from ibeatles.utilities.file_handler import FileHandler

TIME_SPECTRA_NAME_FORMAT = "*_Spectra.txt"


class GetTimeSpectraFilename:
    __slots__ = [
        "parent",
        "file_found",
        "time_spectra",
        "time_spectra_name_format",
        "folder",
    ]

    def __init__(self, parent=None, folder=None):
        self.parent = parent
        self.file_found = False
        self.time_spectra = ""
        self.time_spectra_name_format = "*_Spectra.txt"
        self.folder = folder

    def retrieve_file_name(self):
        time_spectra = glob.glob(self.folder + "/" + TIME_SPECTRA_NAME_FORMAT)
        if time_spectra and os.path.exists(time_spectra[0]):
            return f"{time_spectra[0]}"

        else:
            return ""


class TimeSpectraHandler:
    tof_array = []
    lambda_array = []
    counts_array = []
    full_file_name = ""

    def __init__(self, parent=None, time_spectra_file_name=None):
        self.tof_array = []
        self.parent = parent

        filename = time_spectra_file_name

        self.short_file_name = Path(filename).name
        self.full_file_name = Path(filename)

    def load(self):
        if self.full_file_name.is_file():
            _tof_handler = TOF(filename=str(self.full_file_name))
            _tof_array_s = _tof_handler.tof_array
            # self.tof_array = _tof_array_s * 1e6
            self.tof_array = _tof_array_s
            self.counts_array = _tof_handler.counts_array

    def calculate_lambda_scale(self):
        distance_source_detector = float(self.parent.ui.distance_source_detector_label.text())
        detector_offset = float(self.parent.ui.detector_offset_label.text())

        _exp = Experiment(
            tof=self.tof_array,
            distance_source_detector_m=distance_source_detector,
            detector_offset_micros=detector_offset,
        )
        self.lambda_array = _exp.lambda_array


class TimeSpectraLauncher:
    def __init__(self, parent=None):
        self.parent = parent

        if self.parent.time_spectra[TimeSpectraKeys.tof_array] is None:
            return

        short_file_name = os.path.basename(self.parent.time_spectra[TimeSpectraKeys.file_name])
        full_file_name = self.parent.time_spectra[TimeSpectraKeys.file_name]
        x_axis = self.parent.time_spectra[TimeSpectraKeys.tof_array]
        y_axis = self.parent.time_spectra[TimeSpectraKeys.counts_array]
        x2_axis = self.parent.time_spectra[TimeSpectraKeys.lambda_array]

        _time_spectra_window = TimeSpectraDisplay(
            parent=self.parent,
            short_filename=short_file_name,
            full_filename=full_file_name,
            x_axis=x_axis,
            y_axis=y_axis,
            x2_axis=x2_axis,
        )
        _time_spectra_window.show()


class TimeSpectraDisplay(QMainWindow):
    def __init__(
        self,
        parent=None,
        short_filename="",
        full_filename="",
        x_axis=[],
        y_axis=[],
        x2_axis=[],
    ):
        self.parent = parent
        self.x_axis = x_axis
        self.x2_axis = x2_axis
        self.y_axis = y_axis
        self.full_filename = full_filename

        QMainWindow.__init__(self, parent=parent)
        self.ui = load_ui("ui_time_spectra_preview.ui", baseinstance=self)

        self.initialize_view()
        self.setWindowTitle(short_filename)
        self.populate_text()
        self.plot_data()

    def initialize_view(self):
        graphics_view_layout = QVBoxLayout()
        self.ui.time_spectra_view.setLayout(graphics_view_layout)
        self.ui.time_spectra_plot = MatplotlibView(self.parent)
        graphics_view_layout.addWidget(self.ui.time_spectra_plot)

    def populate_text(self):
        _file_contain = FileHandler.retrieve_ascii_contain(self.full_filename)
        self.ui.time_spectra_text.setText(_file_contain)

    def plot_data(self):
        self.ui.time_spectra_plot.ax1.plot(self.x_axis, self.y_axis, ".")

        # if not self.x2_axis == []:
        #     ax2 = self.ui.time_spectra_plot.canvas.ax.twiny()
        #     ax2.plot(self.x2_axis, np.ones(len(self.x2_axis)), '.')
        #     ax2.cla()
        #     ax2.set_xlabel(r"$Lambda  (\AA)$")

        self.ui.time_spectra_plot.ax1.set_xlabel(r"$TOF  (\mu s)$")
        self.ui.time_spectra_plot.ax1.set_ylabel("Counts")
        self.ui.time_spectra_plot.figure.subplots_adjust(top=0.9, left=0.1)

        self.ui.time_spectra_plot.draw()


class MatplotlibView(FigureCanvas):
    def __init__(self, parent):
        self.fig = Figure()
        self.ax1 = self.fig.add_subplot(111)
        FigureCanvas.__init__(self, self.fig)
        self.setParent(parent)

        FigureCanvas.setSizePolicy(self, QSizePolicy.Expanding, QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)



#!/usr/bin/env python
"""
LoadBin class for loading binning tab data from session file.
"""

import numpy as np

from ibeatles import DataType
from ibeatles.session import SessionKeys, SessionSubKeys


class LoadBin:
    def __init__(self, parent=None):
        self.parent = parent
        self.session_dict = parent.session_dict

    def all(self):
        session_dict = self.session_dict

        # self.parent.data_metadata[DataType.bin]['ui_accessed'] = session_dict[DataType.bin]['ui accessed']
        # self.parent.binning_roi = session_dict[SessionKeys.bin][SessionSubKeys.roi]
        self.parent.there_is_a_roi = True

        self.parent.image_view_settings[DataType.bin]["state"] = session_dict[SessionKeys.bin][
            SessionSubKeys.image_view_state
        ]
        self.parent.image_view_settings[DataType.bin]["histogram"] = session_dict[SessionKeys.bin][
            SessionSubKeys.image_view_histogram
        ]

        binning_line_view = session_dict[SessionKeys.bin][SessionSubKeys.binning_line_view]
        self.parent.binning_line_view["pos"] = np.array(binning_line_view["pos"])
        self.parent.binning_line_view["adj"] = np.array(binning_line_view["adj"])

        if self.parent.binning_line_view["pos"] is None:
            line_color = tuple(binning_line_view["line color"])
            lines = np.array(
                [line_color for n in np.arange(len(self.parent.binning_line_view["pos"]))],
                dtype=[
                    ("red", np.ubyte),
                    ("green", np.ubyte),
                    ("blue", np.ubyte),
                    ("alpha", np.ubyte),
                    ("width", float),
                ],
            )
            self.parent.binning_line_view["pen"] = lines



#!/usr/bin/env python
"""
AutoEventHandler class
"""

import logging

import numpy as np
import pyqtgraph as pg
from qtpy import QtGui
from qtpy.QtWidgets import QCheckBox, QMenu

# from . import TO_MICROS_UNITS, TO_ANGSTROMS_UNITS
from ibeatles.tools.tof_bin import TO_ANGSTROMS_UNITS, TO_MICROS_UNITS, BinAutoMode
from ibeatles.tools.tof_bin.linear_bin import LinearBin
from ibeatles.tools.tof_bin.log_bin import LogBin
from ibeatles.tools.tof_bin.plot import Plot
from ibeatles.tools.tof_bin.utilities.get import Get
from ibeatles.tools.utilities import TimeSpectraKeys
from ibeatles.utilities.status_message_config import (
    StatusMessageStatus,
    show_status_message,
)

# from .log_bin import LogBin
from ibeatles.utilities.table_handler import TableHandler

# from .plot import Plot
# from ..utilities.table_handler import TableHandler
# from ..utilities.status_message_config import StatusMessageStatus, show_status_message
# from .linear_bin import LinearBin

FILE_INDEX_BIN_MARGIN = 0.5
UNSELECTED_BIN = (0, 0, 200, 50)
SELECTED_BIN = (0, 200, 0, 50)


class AutoEventHandler:
    def __init__(self, parent=None):
        self.parent = parent
        self.logger = logging.getLogger("maverick")

        self.tof_bin_margin = (
            self.parent.time_spectra[TimeSpectraKeys.tof_array][1]
            - self.parent.time_spectra[TimeSpectraKeys.tof_array][0]
        ) / 2.0

        self.lambda_bin_margin = (
            self.parent.time_spectra[TimeSpectraKeys.lambda_array][1]
            - self.parent.time_spectra[TimeSpectraKeys.lambda_array][0]
        ) / 2

    def refresh_auto_tab(self):
        # refresh the right plot with profile + bin selected when the auto tab is selected
        o_plot = Plot(parent=self.parent)
        o_plot.refresh_profile_plot_and_clear_bins()

        o_get = Get(parent=self.parent)
        time_spectra_x_axis_name = o_get.x_axis_selected()

        if o_get.bin_auto_mode() == BinAutoMode.linear:
            bins = self.parent.linear_bins[time_spectra_x_axis_name]
            bins_selected = self.parent.linear_bins_selected
        else:
            bins = self.parent.log_bins[time_spectra_x_axis_name]
            bins_selected = self.parent.log_bins_selected

        dict_of_bins_item = {}
        for _index, _bin in enumerate(bins):
            if _bin == []:
                continue

            if time_spectra_x_axis_name == TimeSpectraKeys.file_index_array:
                scale_bin = [
                    _bin[0] - FILE_INDEX_BIN_MARGIN,
                    _bin[-1] + FILE_INDEX_BIN_MARGIN,
                ]

            elif time_spectra_x_axis_name == TimeSpectraKeys.tof_array:
                scale_bin = [
                    _bin[0] - self.tof_bin_margin,
                    _bin[-1] + self.tof_bin_margin,
                ]
                scale_bin = [_value * TO_MICROS_UNITS for _value in scale_bin]

            else:
                scale_bin = [
                    _bin[0] - self.lambda_bin_margin,
                    _bin[-1] + self.lambda_bin_margin,
                ]
                scale_bin = [_value * TO_ANGSTROMS_UNITS for _value in scale_bin]

            item = pg.LinearRegionItem(
                values=scale_bin,
                orientation="vertical",
                brush=UNSELECTED_BIN,
                movable=False,
                bounds=None,
            )
            item.setZValue(-10)
            self.parent.bin_profile_view.addItem(item)
            dict_of_bins_item[_index] = item

        self.parent.dict_of_bins_item = dict_of_bins_item

        # bring back the row selected
        if bins_selected:
            o_table = TableHandler(table_ui=self.parent.ui.bin_auto_tableWidget)
            o_table.select_rows(bins_selected)

    def auto_linear_radioButton_changed(self):
        file_index_status = False
        tof_status = False
        lambda_status = False
        if self.parent.ui.auto_linear_file_index_radioButton.isChecked():
            file_index_status = True
            source_button = TimeSpectraKeys.file_index_array
        elif self.parent.ui.auto_linear_tof_radioButton.isChecked():
            tof_status = True
            source_button = TimeSpectraKeys.tof_array
        else:
            lambda_status = True
            source_button = TimeSpectraKeys.lambda_array

        self.parent.ui.auto_linear_file_index_spinBox.setEnabled(file_index_status)
        self.parent.ui.auto_linear_tof_doubleSpinBox.setEnabled(tof_status)
        self.parent.ui.bin_auto_linear_tof_units_label.setEnabled(tof_status)
        self.parent.ui.auto_linear_lambda_doubleSpinBox.setEnabled(lambda_status)
        self.parent.ui.bin_auto_linear_lambda_units_label.setEnabled(lambda_status)

        self.bin_auto_linear_changed(source_radio_button=source_button)

    def auto_log_radioButton_changed(self):
        file_index_status = False
        tof_status = False
        lambda_status = False
        if self.parent.ui.bin_auto_log_file_index_radioButton.isChecked():
            file_index_status = True
            source_button = TimeSpectraKeys.file_index_array
        elif self.parent.ui.bin_auto_log_tof_radioButton.isChecked():
            tof_status = True
            source_button = TimeSpectraKeys.tof_array
        else:
            lambda_status = True
            source_button = TimeSpectraKeys.lambda_array

        self.parent.ui.auto_log_file_index_spinBox.setEnabled(file_index_status)
        self.parent.ui.auto_log_tof_doubleSpinBox.setEnabled(tof_status)
        self.parent.ui.auto_log_lambda_doubleSpinBox.setEnabled(lambda_status)

        self.bin_auto_log_changed(source_radio_button=source_button)

    def bin_auto_radioButton_clicked(self):
        state_auto = self.parent.ui.auto_log_radioButton.isChecked()
        self.parent.ui.auto_linear_radioButton.setChecked(state_auto)
        self.parent.ui.auto_linear_radioButton.setChecked(not state_auto)
        self.parent.ui.bin_auto_log_frame.setEnabled(state_auto)
        self.parent.ui.bin_auto_linear_frame.setEnabled(not state_auto)
        o_get = Get(parent=self.parent)
        if o_get.bin_auto_mode() == BinAutoMode.log:
            self.auto_log_radioButton_changed()
        else:
            self.auto_linear_radioButton_changed()

    def bin_auto_log_changed(self, source_radio_button=TimeSpectraKeys.file_index_array):
        self.logger.info(f"bin auto log changed: radio button changed -> {source_radio_button}")
        o_bin = LogBin(parent=self.parent, source_radio_button=source_radio_button)

        self.parent.ui.auto_log_file_index_spinBox.blockSignals(True)
        self.parent.ui.auto_log_tof_doubleSpinBox.blockSignals(True)
        self.parent.ui.auto_log_lambda_doubleSpinBox.blockSignals(True)

        self.logger.info(
            f"-> original raw_file_index_array_binned: {self.parent.time_spectra[TimeSpectraKeys.file_index_array]}"
        )
        self.logger.info(f"-> original raw_tof_array_binned: {self.parent.time_spectra[TimeSpectraKeys.tof_array]}")
        self.logger.info(
            f"-> original raw_lambda_array_binned: {self.parent.time_spectra[TimeSpectraKeys.lambda_array]}"
        )

        o_get = Get(parent=self.parent)
        log_bin_requested = o_get.auto_log_bin_requested()
        self.logger.info(f"--> bin requested: {log_bin_requested}")

        if source_radio_button == TimeSpectraKeys.file_index_array:
            o_bin.create_log_file_index_bin_array(bin_value=log_bin_requested)

        elif source_radio_button == TimeSpectraKeys.tof_array:
            o_bin.create_log_file_index_bin_array(bin_value=log_bin_requested)
            o_bin.create_log_bin_arrays()

        elif source_radio_button == TimeSpectraKeys.lambda_array:
            o_bin.create_log_file_index_bin_array(bin_value=log_bin_requested)
            o_bin.create_log_bin_arrays()

        else:
            raise NotImplementedError("bin auto log algorithm not implemented!")

        self.logger.info(f"-> file_index_array_binned: {o_bin.log_bins[TimeSpectraKeys.file_index_array]}")
        self.logger.info(f"-> tof_array_binned: {o_bin.log_bins[TimeSpectraKeys.tof_array]}")
        self.logger.info(f"-> lambda_array_binned: {o_bin.log_bins[TimeSpectraKeys.lambda_array]}")

        self.parent.log_bins = {
            TimeSpectraKeys.file_index_array: o_bin.get_log_file_index(),
            TimeSpectraKeys.tof_array: o_bin.get_log_tof(),
            TimeSpectraKeys.lambda_array: o_bin.get_log_lambda(),
        }

        self.fill_auto_table()
        self.update_auto_table()
        self.refresh_auto_tab()

        self.parent.ui.auto_log_file_index_spinBox.blockSignals(False)
        self.parent.ui.auto_log_tof_doubleSpinBox.blockSignals(False)
        self.parent.ui.auto_log_lambda_doubleSpinBox.blockSignals(False)

    def fill_auto_table(self):
        o_table = TableHandler(table_ui=self.parent.ui.bin_auto_tableWidget)
        o_table.remove_all_rows()

        o_get = Get(parent=self.parent)
        bin_auto_mode = o_get.bin_auto_mode()

        if bin_auto_mode == BinAutoMode.linear:
            bins = self.parent.linear_bins
        else:
            bins = self.parent.log_bins

        list_rows = np.arange(len(bins[TimeSpectraKeys.file_index_array]))

        file_index_array_of_bins = bins[TimeSpectraKeys.file_index_array]
        tof_array_of_bins = bins[TimeSpectraKeys.tof_array]
        lambda_array_of_bins = bins[TimeSpectraKeys.lambda_array]

        for _row in np.arange(len(list_rows)):
            file_bin = file_index_array_of_bins[_row]
            tof_bin = tof_array_of_bins[_row]
            lambda_bin = lambda_array_of_bins[_row]

            checkbox_enabled = True
            if not file_bin:
                str_file_index = "N/A"
                str_tof = "N/A"
                str_lambda = "N/A"
                checkbox_enabled = False

            elif len(file_bin) == 1:
                str_file_index = file_bin[0]
                str_tof = f"{tof_bin[0] * TO_MICROS_UNITS: .2f}"
                str_lambda = f"{lambda_bin[0] * TO_ANGSTROMS_UNITS: .3f}"

            elif len(file_bin) == 2:
                str_file_index = f"{file_bin[0]}, {file_bin[1]}"
                str_tof = f"{tof_bin[0] * TO_MICROS_UNITS:.2f}, {tof_bin[1] * TO_MICROS_UNITS:.2f}"
                str_lambda = f"{lambda_bin[0] * TO_ANGSTROMS_UNITS:.3f}, {lambda_bin[1] * TO_ANGSTROMS_UNITS:.3f}"

            else:
                str_file_index = f"{file_bin[0]} ... {file_bin[-1]}"
                str_tof = f"{tof_bin[0] * TO_MICROS_UNITS:.2f} ... {tof_bin[-1] * TO_MICROS_UNITS:.2f}"
                str_lambda = f"{lambda_bin[0] * TO_ANGSTROMS_UNITS:.3f} ... {lambda_bin[-1] * TO_ANGSTROMS_UNITS:.3f}"

            o_table.insert_empty_row(row=_row)

            # use or not that bin
            if checkbox_enabled:
                checkbox = QCheckBox()
                checkbox.setChecked(True)
                checkbox.stateChanged.connect(
                    lambda state=0, row=_row: self.parent.auto_table_use_checkbox_changed(state, row)
                )
                o_table.insert_widget(row=_row, column=0, widget=checkbox, centered=True)

            o_table.insert_item(row=_row, column=1, value=_row, editable=False)
            o_table.insert_item(row=_row, column=2, value=str_file_index, editable=False)
            o_table.insert_item(row=_row, column=3, value=str_tof, editable=False)
            o_table.insert_item(row=_row, column=4, value=str_lambda, editable=False)

    def bin_auto_linear_changed(self, source_radio_button=TimeSpectraKeys.file_index_array):
        self.logger.info(f"bin auto linear changed: radio button changed -> {source_radio_button}")
        o_bin = LinearBin(parent=self.parent, source_array=source_radio_button)
        self.parent.ui.auto_linear_file_index_spinBox.blockSignals(True)
        self.parent.ui.auto_linear_tof_doubleSpinBox.blockSignals(True)
        self.parent.ui.auto_linear_lambda_doubleSpinBox.blockSignals(True)

        self.logger.info(
            f"-> raw_file_index_array_binned: {self.parent.time_spectra[TimeSpectraKeys.file_index_array]}"
        )
        self.logger.info(f"-> raw_tof_array_binned: {self.parent.time_spectra[TimeSpectraKeys.tof_array]}")
        self.logger.info(f"-> raw_lambda_array_binned: {self.parent.time_spectra[TimeSpectraKeys.lambda_array]}")

        if source_radio_button == TimeSpectraKeys.file_index_array:
            file_index_value = self.parent.ui.auto_linear_file_index_spinBox.value()
            self.logger.info(f"--> bin requested: {file_index_value}")
            o_bin.create_linear_file_index_bin_array(bin_value=file_index_value)
            o_bin.create_linear_bin_arrays()

        elif source_radio_button == TimeSpectraKeys.tof_array:
            tof_value = self.parent.ui.auto_linear_tof_doubleSpinBox.value()
            self.logger.info(f"--> bin requested: {tof_value}")
            o_bin.create_linear_file_index_bin_array(bin_value=tof_value * 1e-6)  # to switch to seconds
            o_bin.create_linear_bin_arrays()

        elif source_radio_button == TimeSpectraKeys.lambda_array:
            lambda_value = self.parent.ui.auto_linear_lambda_doubleSpinBox.value()
            self.logger.info(f"--> bin requested: {lambda_value}")
            o_bin.create_linear_file_index_bin_array(bin_value=lambda_value * 1e-10)  # to switch to seconds
            o_bin.create_linear_bin_arrays()

        else:
            raise NotImplementedError("bin auto linear algorithm not implemented!")

        self.logger.info(f"-> file_index_array_binned: {o_bin.linear_bins[TimeSpectraKeys.file_index_array]}")
        self.logger.info(f"-> tof_array_binned: {o_bin.linear_bins[TimeSpectraKeys.tof_array]}")
        self.logger.info(f"-> lambda_array_binned: {o_bin.linear_bins[TimeSpectraKeys.lambda_array]}")

        self.parent.linear_bins = {
            TimeSpectraKeys.file_index_array: o_bin.get_linear_file_index(),
            TimeSpectraKeys.tof_array: o_bin.get_linear_tof(),
            TimeSpectraKeys.lambda_array: o_bin.get_linear_lambda(),
        }

        self.fill_auto_table()
        self.refresh_auto_tab()

        show_status_message(
            parent=self.parent,
            message=f"New {source_radio_button} bin size selected!",
            status=StatusMessageStatus.ready,
            duration_s=5,
        )

        self.parent.ui.auto_linear_file_index_spinBox.blockSignals(False)
        self.parent.ui.auto_linear_tof_doubleSpinBox.blockSignals(False)
        self.parent.ui.auto_linear_lambda_doubleSpinBox.blockSignals(False)

    def update_auto_table(self):
        o_table = TableHandler(table_ui=self.parent.ui.bin_auto_tableWidget)
        nbr_rows = o_table.row_count()

        if self.parent.ui.bin_auto_hide_empty_bins_checkBox.isChecked():
            for _row in np.arange(nbr_rows):
                item = o_table.get_item_str_from_cell(row=_row, column=2)
                if item == "N/A":
                    o_table.set_row_hidden(_row, True)

        else:
            for _row in np.arange(nbr_rows):
                o_table.set_row_hidden(_row, False)

    def use_auto_bin_state_changed(self, row=0, state=True):
        """
        user change the state of any of the bin checkbox
        :param row:
        :param state: True or False
        """
        o_table = TableHandler(table_ui=self.parent.ui.bin_auto_tableWidget)
        rows_selected = o_table.get_rows_of_table_selected()

        if row in rows_selected:
            for _row in rows_selected:
                item = self.parent.dict_of_bins_item.get(_row, None)
                if item:
                    if state:
                        self.parent.bin_profile_view.addItem(item)
                    else:
                        self.parent.bin_profile_view.removeItem(item)

                    widget = o_table.get_widget(row=_row, column=0)
                    if widget:
                        checkbox = widget.children()[1]
                        if state:
                            checkbox.setChecked(2)
                        else:
                            checkbox.setChecked(0)

        else:
            item = self.parent.dict_of_bins_item.get(row, None)
            if item:
                if state:
                    self.parent.bin_profile_view.addItem(item)
                else:
                    self.parent.bin_profile_view.removeItem(item)

    def auto_table_right_click(self, position=None):
        menu = QMenu(self.parent)

        select_all = menu.addAction("Select all bins")
        unselect_all = menu.addAction("Unselect all bins")

        action = menu.exec_(QtGui.QCursor.pos())
        if action == select_all:
            self.all_auto_bins_checkbox(state=True)
        elif action == unselect_all:
            self.all_auto_bins_checkbox(state=False)
        else:
            pass

    def all_auto_bins_checkbox(self, state=True):
        o_table = TableHandler(table_ui=self.parent.ui.bin_auto_tableWidget)
        nbr_rows = o_table.row_count()
        for _row in np.arange(nbr_rows):
            widget = o_table.get_widget(row=_row, column=0)
            if widget:
                checkbox = widget.children()[1]
                checkbox.setChecked(state)
                self.use_auto_bin_state_changed(row=_row, state=state)

    def use_this_bin(self, row=0):
        o_table = TableHandler(table_ui=self.parent.ui.bin_auto_tableWidget)
        widget = o_table.get_widget(row=row, column=0)
        if widget:
            checkbox = widget.children()[1]
            return checkbox.isChecked()
        else:
            return False

    def auto_table_selection_changed(self):
        o_get = Get(parent=self.parent)
        if o_get.bin_auto_mode() == BinAutoMode.linear:
            previous_rows_highlighted = self.parent.linear_bins_selected
        else:
            previous_rows_highlighted = self.parent.log_bins_selected

        if previous_rows_highlighted:
            for _row in previous_rows_highlighted:
                previous_item = self.parent.dict_of_bins_item.get(_row, None)
                if previous_item:
                    self.parent.bin_profile_view.removeItem(previous_item)
                    previous_item.setBrush(pg.mkBrush(UNSELECTED_BIN))
                    if self.use_this_bin(row=_row):
                        self.parent.bin_profile_view.addItem(previous_item)

        o_table = TableHandler(table_ui=self.parent.ui.bin_auto_tableWidget)
        new_rows_to_highlight = o_table.get_rows_of_table_selected()
        clean_list_of_new_rows_to_highlight = []  # removing rows that can not be selected (empty bins)
        for _row in new_rows_to_highlight:
            new_item = self.parent.dict_of_bins_item.get(_row, None)
            if new_item:
                self.parent.bin_profile_view.removeItem(new_item)
                new_item.setBrush(pg.mkBrush(SELECTED_BIN))
                self.parent.bin_profile_view.addItem(new_item)
                clean_list_of_new_rows_to_highlight.append(_row)

        self.parent.current_auto_bin_rows_highlighted = clean_list_of_new_rows_to_highlight

        if o_get.bin_auto_mode() == BinAutoMode.linear:
            self.parent.linear_bins_selected = clean_list_of_new_rows_to_highlight
        else:
            self.parent.log_bins_selected = clean_list_of_new_rows_to_highlight

    def clear_selection(self, auto_mode=BinAutoMode.linear):
        if auto_mode == BinAutoMode.linear:
            self.parent.linear_bins_selected = None
        else:
            self.parent.log_bins_selected = None
        o_table = TableHandler(table_ui=self.parent.ui.bin_auto_tableWidget)
        o_table.select_everything(False)



"""Functions for generating bins and extracting bin data for fitting."""

from typing import List, Optional, Sequence, Tuple

import numpy as np

from ibeatles.core.config import BinCoordinates


def get_bin_coordinates(
    image_shape: Tuple[int, int],
    x0: int,
    y0: int,
    width: int,
    height: int,
    bins_size: int,
) -> List[BinCoordinates]:
    """
    Generate list of bin coordinates based on configuration.

    Parameters
    ----------
    image_shape : Tuple[int, int]
        Shape of the image (height, width)
    x0 : int
        Starting x coordinate of ROI
    y0 : int
        Starting y coordinate of ROI
    width : int
        Width of ROI
    height : int
        Height of ROI
    bins_size : int
        Size of each bin (square bins)

    Returns
    -------
    List[BinCoordinates]
        List of bin coordinates with row and column indices

    Raises
    ------
    ValueError
        If ROI extends beyond image boundaries or bins_size is invalid
    """
    # Validate inputs
    img_height, img_width = image_shape

    if x0 < 0 or y0 < 0:
        raise ValueError("Starting coordinates must be non-negative")

    if x0 + width > img_width or y0 + height > img_height:
        raise ValueError("ROI extends beyond image boundaries")

    if bins_size <= 0:
        raise ValueError("Bin size must be positive")

    if width < bins_size or height < bins_size:
        raise ValueError("Bin size larger than ROI dimensions")

    # Calculate number of bins in each direction
    n_cols = width // bins_size  # Integer division
    n_rows = height // bins_size

    bins = []
    for col in range(n_cols):
        for row in range(n_rows):
            bin_x0 = x0 + col * bins_size
            bin_x1 = bin_x0 + bins_size
            bin_y0 = y0 + row * bins_size
            bin_y1 = bin_y0 + bins_size

            # Create bin coordinates with grid position
            bin_coords = BinCoordinates(
                x0=bin_x0,
                x1=bin_x1,
                y0=bin_y0,
                y1=bin_y1,
                row_index=row,
                column_index=col,
            )

            bins.append(bin_coords)

    return bins


def get_bin_transmission(
    images: Sequence[np.ndarray],
    wavelengths: np.ndarray,
    bin_coords: BinCoordinates,
    lambda_range: Optional[Tuple[float, float]] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Extract wavelength-transmission profile for a single bin.

    Parameters
    ----------
    images : Sequence[np.ndarray]
        List or array of 2D transmission images
    wavelengths : np.ndarray
        Corresponding wavelengths for each image
    bin_coords : BinCoordinates
        Coordinates of the bin
    lambda_range : Optional[Tuple[float, float]]
        Optional (min, max) wavelength range to extract. If None, use all wavelengths.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        Wavelengths and corresponding transmission values for the bin

    Notes
    -----
    - Uses numpy.nanmean to handle any NaN values in the data
    - Assumes images are already normalized to transmission values
    """
    # Extract bin region from each image and calculate mean
    # NOTE: some additional filtering might be helpful here is the data
    #       is extremely noisy
    transmission_values = np.array(
        [np.nanmean(img[bin_coords.y0 : bin_coords.y1, bin_coords.x0 : bin_coords.x1]) for img in images]
    )

    # Apply wavelength range selection if specified
    if lambda_range is not None:
        lambda_min, lambda_max = lambda_range
        mask = (wavelengths >= lambda_min) & (wavelengths <= lambda_max)
        wavelengths = wavelengths[mask]
        transmission_values = transmission_values[mask]

    return wavelengths, transmission_values


def validate_transmission_data(wavelengths: np.ndarray, transmission: np.ndarray, min_valid_points: int = 10) -> bool:
    """
    Validate transmission data before fitting.

    Parameters
    ----------
    wavelengths : np.ndarray
        Wavelength values
    transmission : np.ndarray
        Transmission values
    min_valid_points : int, optional
        Minimum number of valid points required, by default 10

    Returns
    -------
    bool
        True if data is valid for fitting

    Notes
    -----
    Checks:
    - Matching wavelength and transmission lengths
    - Sufficient number of valid (non-NaN) points
    - All transmission values are between 0 and 1
    - Wavelengths are monotonically increasing
    """
    # Check lengths match
    if len(wavelengths) != len(transmission):
        return False

    # Check number of valid points
    valid_mask = ~(np.isnan(transmission) | np.isnan(wavelengths))
    if np.sum(valid_mask) < min_valid_points:
        return False

    # Check transmission values are physical
    valid_transmission = transmission[valid_mask]
    if not np.all((valid_transmission >= 0) & (valid_transmission <= 1)):
        return False

    # Check wavelengths are ordered
    valid_wavelengths = wavelengths[valid_mask]
    if not np.all(np.diff(valid_wavelengths) > 0):
        return False

    return True



import os

from qtpy.uic import loadUi


def load_ui(ui_filename, baseinstance):
    ui_filename = os.path.split(ui_filename)[-1]
    ui_path = os.path.join(os.path.dirname(__file__), "..", "ui")
    filename = os.path.join(ui_path, ui_filename)
    return loadUi(filename, baseinstance=baseinstance)



#!/usr/bin/env python
"""
Plotting handler for step 2
"""

import numpy as np
import pyqtgraph as pg
from neutronbraggedge.experiment_handler.experiment import Experiment
from qtpy import QtGui
from qtpy.QtWidgets import QCheckBox, QComboBox, QTableWidgetItem

from ibeatles import DataType, RegionType
from ibeatles.step2 import roi_label_color
from ibeatles.step2.get import Get as Step2Get
from ibeatles.utilities.colors import pen_color
from ibeatles.utilities.gui_handler import GuiHandler
from ibeatles.utilities.pyqrgraph import Pyqtgrah as PyqtgraphUtilities


class CustomAxis(pg.AxisItem):
    def __init__(self, gui_parent, *args, **kwargs):
        pg.AxisItem.__init__(self, *args, **kwargs)
        self.parent = gui_parent

    def tickStrings(self, values, scale, spacing):
        strings = []

        _distance_source_detector = float(str(self.parent.ui.distance_source_detector.text()))
        _detector_offset_micros = float(str(self.parent.ui.detector_offset.text()))

        tof_s = [float(time) * 1e-6 for time in values]

        _exp = Experiment(
            tof=tof_s,
            distance_source_detector_m=_distance_source_detector,
            detector_offset_micros=_detector_offset_micros,
        )
        lambda_array = _exp.lambda_array

        for _lambda in lambda_array:
            strings.append("{:.4f}".format(_lambda * 1e10))

        return strings


class Step2Plot:
    sample: list = []
    ob: list = []
    normalization: list = []

    def __init__(self, parent=None, sample=[], ob=[], normalized=[]):
        self.parent = parent
        self.sample = sample
        self.ob = ob
        self.normalized = normalized

    def prepare_data(self):
        if len(self.sample) == 0:
            sample = self.parent.data_metadata["sample"]["data"]

        # still no sample data
        if len(sample) == 0:
            return

        if len(self.ob) == 0:
            ob = self.parent.data_metadata["ob"]["data"]

        if len(self.normalized) == 0:
            normalized = self.parent.data_metadata["normalized"]["data"]

        if len(self.parent.data_metadata["normalization"]["data"]) == 0:
            normalization = np.mean(np.array(sample), axis=0)
            self.parent.data_metadata["normalization"]["axis"] = normalization
            self.parent.data_metadata["normalization"]["data"] = normalization
        else:
            normalization = self.parent.data_metadata["normalization"]["data"]

        self.sample = sample
        self.ob = ob
        self.normalization = normalization
        self.normalized = normalized

    def clear_image(self):
        self.parent.step2_ui["image_view"].clear()

    def clear_roi_from_image(self):
        list_roi_id = self.parent.list_roi_id["normalization"]
        for roi_id in list_roi_id:
            self.parent.step2_ui["image_view"].removeItem(roi_id)

    def display_image(self):
        _data = self.normalization

        o_pyqt = PyqtgraphUtilities(
            parent=self.parent,
            image_view=self.parent.step2_ui["image_view"],
            data_type=DataType.normalization,
        )
        _state = o_pyqt.get_state()
        o_pyqt.save_histogram_level()

        if len(_data) == 0:
            self.clear_plots()
            self.parent.step2_ui["area"].setVisible(False)
        else:
            self.parent.step2_ui["area"].setVisible(True)
            self.parent.step2_ui["image_view"].setImage(_data)

        o_pyqt.set_state(_state)
        o_pyqt.reload_histogram_level()

    def display_roi(self):
        list_roi_id = self.parent.list_roi_id["normalization"]
        list_label_roi_id = self.parent.list_label_roi_id["normalization"]
        roi = self.parent.list_roi["normalization"]

        if len(list_roi_id) == 0:
            return

        self.clear_roi_from_image()

        for index, roi_id in enumerate(list_roi_id):
            [_, x0, y0, width, height, region_type] = roi[index]

            x0 = int(x0)
            y0 = int(y0)

            roi_id.setPos([x0, y0], update=False, finish=False)
            roi_id.setSize([width, height], update=False, finish=False)

            _label_roi_id = list_label_roi_id[index]
            _label_roi_id.setPos(x0, y0)
            _label_roi_id.setHtml(
                '<div style="text-align: center"><span style="color: #ff0000;">' + region_type + "</span></div>"
            )

    def display_bragg_edge(self):
        def set_curve_point(text=RegionType.sample, parent_curve=None):
            curve_point = pg.CurvePoint(parent_curve)
            _plot_ui.addItem(curve_point)
            _text = pg.TextItem(text, anchor=(0.5, 0))
            _text.setParentItem(curve_point)
            arrow = pg.ArrowItem(angle=0)
            arrow.setParentItem(curve_point)
            curve_point.setPos(x_axis[-1])

        _plot_ui = self.parent.step2_ui["bragg_edge_plot"]
        _plot_ui.clear()

        list_roi_id = self.parent.list_roi_id["normalization"]
        # list_roi = self.parent.list_roi['normalization']

        o_get = Step2Get(parent=self.parent)
        list_sample_roi = []
        list_background_roi = []
        for _row_index, roi in enumerate(list_roi_id):
            [flag, x0, y0, width, height, region_type] = o_get.roi_table_row(row=_row_index)
            if flag is False:
                continue

            if region_type == RegionType.sample:
                list_sample_roi.append([x0, y0, width, height])
            else:
                list_background_roi.append([x0, y0, width, height])

        data_to_plot = self.extract_data_from_roi(
            list_sample_roi=list_sample_roi, list_background_roi=list_background_roi
        )

        if data_to_plot is None:
            return

        if (not data_to_plot[RegionType.sample]) and (not data_to_plot[RegionType.background]):
            return

        o_gui = GuiHandler(parent=self.parent)
        xaxis_choice = o_gui.get_step2_xaxis_checked()
        if xaxis_choice == "file_index":
            if data_to_plot[RegionType.sample]:
                x_axis = np.arange(len(data_to_plot[RegionType.sample]))
            else:
                x_axis = np.arange(len(data_to_plot[RegionType.background]))
            _plot_ui.setLabel("bottom", "File Index")

        elif xaxis_choice == "tof":
            tof_array = self.parent.data_metadata["time_spectra"]["data"]
            tof_array = tof_array * 1e6
            _plot_ui.setLabel("bottom", "TOF (\u00b5s)")
            x_axis = tof_array

        else:
            lambda_array = self.parent.data_metadata["time_spectra"]["lambda"]
            lambda_array = lambda_array * 1e10
            _plot_ui.setLabel("bottom", "\u03bb (\u212b)")
            x_axis = lambda_array

        if data_to_plot[RegionType.sample]:
            # display the profile for the sample
            curve = _plot_ui.plot(
                x_axis,
                data_to_plot[RegionType.sample],
                symbolPen=None,
                pen=pen_color["0"],
                symbol="t",
                symbolSize=5,
            )
            set_curve_point(text=RegionType.sample, parent_curve=curve)

        if data_to_plot[RegionType.background]:
            # display the profile for the background
            curve = _plot_ui.plot(
                x_axis,
                data_to_plot[RegionType.background],
                symbolPen=None,
                pen=pen_color["1"],
                symbol="t",
                symbolSize=5,
            )
            set_curve_point(text=RegionType.background, parent_curve=curve)

    def extract_data_from_roi(self, list_sample_roi=None, list_background_roi=None):
        data_to_plot = {RegionType.sample: None, RegionType.background: None}

        data = self.parent.data_metadata[DataType.sample]["data"]

        if list_sample_roi:
            sample_y_profile = []
            for _index_data, _data in enumerate(data):
                total_counts = 0
                total_pixel = 0
                for _roi in list_sample_roi:
                    [x0, y0, width, height] = _roi

                    x0 = int(x0)
                    y0 = int(y0)
                    width = int(width)
                    height = int(height)

                    total_counts += np.sum(_data[x0 : x0 + width, y0 : y0 + height])
                    total_pixel += width * height

                sample_y_profile.append(total_counts / total_pixel)

            data_to_plot[RegionType.sample] = sample_y_profile

        if list_background_roi:
            background_y_profile = []
            for _data in data:
                total_counts = 0
                total_pixel = 0
                for _roi in list_background_roi:
                    [x0, y0, width, height] = _roi

                    x0 = int(x0)
                    y0 = int(y0)
                    width = int(width)
                    height = int(height)

                    total_counts += np.sum(_data[x0 : x0 + width, y0 : y0 + height])
                    total_pixel += width * height

                if total_pixel == 0:
                    return None

                background_y_profile.append(total_counts / total_pixel)

            data_to_plot[RegionType.background] = background_y_profile

        return data_to_plot

    def calculate_mean_counts(self, data, list_roi=None):
        if len(data) == 0:
            return data

        data = np.array(data)
        final_array = []
        _first_array_added = True
        nbr_roi = len(list_roi)

        if len(list_roi) == 0:
            final_array = np.mean(data, axis=(1, 2))

        else:
            for _index, _roi in enumerate(list_roi):
                [x0, y0, width, height] = _roi

                _x_from = int(x0)
                _x_to = _x_from + int(width) + 1

                _y_from = int(y0)
                _y_to = _y_from + int(height) + 1

                _mean = np.mean(data[:, _x_from:_x_to, _y_from:_y_to], axis=(1, 2))
                if _first_array_added:
                    final_array = _mean
                    _first_array_added = False
                else:
                    final_array += _mean

            final_array /= nbr_roi

        return final_array

    def init_roi_table(self):
        # clear table
        for _row in np.arange(self.parent.ui.normalization_tableWidget.rowCount()):
            self.parent.ui.normalization_tableWidget.removeRow(0)

        list_roi = self.parent.list_roi["normalization"]
        self.parent.ui.normalization_tableWidget.blockSignals(True)
        for _row, _roi in enumerate(list_roi):
            self.parent.ui.normalization_tableWidget.insertRow(_row)
            self.set_row(_row, _roi)
        self.parent.ui.normalization_tableWidget.blockSignals(False)

    def update_label_roi(self):
        list_roi = self.parent.list_roi["normalization"]
        list_label_roi_id = self.parent.list_label_roi_id["normalization"]

        for _row, _roi in enumerate(list_roi):
            [status_row, x0, y0, width, height, region_type] = _roi
            x0 = int(x0)
            y0 = int(y0)
            _label_roi_id = list_label_roi_id[_row]
            _label_roi_id.setPos(x0, y0)
            _label_roi_id.setHtml(
                f'<div style="text-align: center"><span style="color: '
                f'{roi_label_color[region_type]};">' + region_type + "</span></div>"
            )

    def check_error_in_roi_table(self):
        list_roi = self.parent.list_roi["normalization"]

        for _row, _roi in enumerate(list_roi):
            [status_row, x0, y0, width, height, region_type] = _roi
            width = int(width)
            height = int(height)

            if (width == 0) or (height == 0):
                are_all_cells_ok = False
            else:
                are_all_cells_ok = True

            self._set_roi_row_background(are_all_cells_ok=are_all_cells_ok, row=_row)

    def _set_roi_row_background(self, are_all_cells_ok=True, row=0):
        if are_all_cells_ok:
            color = QtGui.QColor(255, 255, 255)
        else:
            color = QtGui.QColor(255, 0, 0)

        for _col in np.arange(1, 5):
            _item = self.parent.ui.normalization_tableWidget.item(row, _col)
            _item.setBackground(color)

    def update_roi_table(self):
        # if self.sample == []:
        #     return

        list_roi = self.parent.list_roi["normalization"]
        for _row, _roi in enumerate(list_roi):
            self.update_row(_row, _roi)

    def get_item(self, text):
        _item = QTableWidgetItem(text)
        return _item

    def set_row(self, row_index, roi_array):
        # region type is either 'sample' or 'background'
        [status_row, x0, y0, width, height, region_type] = roi_array

        # button
        _widget = QCheckBox()
        _widget.setChecked(bool(status_row))
        self.parent.ui.normalization_tableWidget.setCellWidget(row_index, 0, _widget)
        self.parent.ui.normalization_tableWidget.blockSignals(True)

        # x0
        _item1 = self.get_item(str(x0))
        self.parent.ui.normalization_tableWidget.setItem(row_index, 1, _item1)

        # y0
        _item2 = self.get_item(str(y0))
        self.parent.ui.normalization_tableWidget.setItem(row_index, 2, _item2)

        # width
        _item3 = self.get_item(str(width))
        self.parent.ui.normalization_tableWidget.setItem(row_index, 3, _item3)

        # height
        _item4 = self.get_item(str(height))
        self.parent.ui.normalization_tableWidget.setItem(row_index, 4, _item4)

        _widget.stateChanged.connect(self.parent.normalization_row_status_changed)
        self.parent.ui.normalization_tableWidget.blockSignals(False)

        # region type
        _widget = QComboBox()
        _widget.addItems([RegionType.sample, RegionType.background])
        index = 0 if (region_type == RegionType.sample) else 1
        _widget.setCurrentIndex(index)
        _widget.currentIndexChanged.connect(self.parent.normalization_row_status_region_type_changed)
        self.parent.ui.normalization_tableWidget.setCellWidget(row_index, 5, _widget)

    def update_row(self, row_index, roi_array):
        [status_row, x0, y0, width, height, _] = roi_array

        # button
        _widget = self.parent.ui.normalization_tableWidget.cellWidget(row_index, 0)
        _widget.setChecked(bool(status_row))

        # x0
        _item = self.parent.ui.normalization_tableWidget.item(row_index, 1)
        _item.setText(str(x0))

        # y0
        _item = self.parent.ui.normalization_tableWidget.item(row_index, 2)
        _item.setText(str(y0))

        # width
        _item = self.parent.ui.normalization_tableWidget.item(row_index, 3)
        _item.setText(str(width))

        # height
        _item = self.parent.ui.normalization_tableWidget.item(row_index, 4)
        _item.setText(str(height))

    def clear_plots(self):
        self.parent.step2_ui["image_view"].clear()
        self.parent.step2_ui["bragg_edge_plot"].clear()
        # self.parent.step2_ui['normalized_profile_plot'].clear()

    def clear_counts_vs_file(self):
        self.parent.step2_ui["bragg_edge_plot"].clear()

    def multiply_array_by_coeff(self, data=None, coeff=None):
        if len(data) == len(coeff):
            return data * coeff
        else:
            return []



import math

import numpy as np
from scipy.special import erfc


def basic_fit(t, d_spacing, alpha, sigma, a1, a2):
    t0 = 2 * d_spacing
    term2 = erfc(-((t - t0) / (sigma * math.sqrt(2))) + sigma / alpha)
    term1 = np.exp((t - t0) / alpha + (sigma * sigma) / (2 * alpha * alpha))
    term0 = erfc(-((t - t0) / (sigma * math.sqrt(2))))

    y = a1 + a2 * (term0 - (term1 * term2))

    return y


def advanced_fit(t, d_spacing, alpha, sigma, a1, a2, a5, a6):
    t0 = 2 * d_spacing
    term0 = a2 * (t - a6)
    term1 = ((a5 - a2) / 2) * (t - a6)
    term3 = erfc(-((t - t0) / (sigma * math.sqrt(2))))
    term4 = np.exp(-((t - t0) / alpha) + ((sigma * sigma) / (2 * alpha * alpha)))
    term5 = erfc(-((t - t0) / (sigma * math.sqrt(2))) + sigma / alpha)

    y = a1 + term0 + term1 * (term3 - term4 * term5)

    return y



from .. import DataType
from ..all_steps.event_handler import EventHandler as GeneralEventHandler
from ..binning.binning_launcher import BinningLauncher
from ..fitting.fitting_launcher import FittingLauncher


class SessionUtilities:
    def __init__(self, parent=None):
        self.parent = parent

    def jump_to_tab_of_data_type(self, data_type=DataType.sample):
        if data_type == DataType.sample:
            return
        elif data_type == DataType.normalization:
            self.parent.ui.tabWidget.setCurrentIndex(1)
        elif data_type == DataType.normalized:
            self.parent.ui.tabWidget.setCurrentIndex(2)
        elif data_type == DataType.bin:
            o_event = GeneralEventHandler(parent=self.parent)
            if o_event.is_step_selected_allowed(step_index_requested=3):
                BinningLauncher(parent=self.parent)
        elif data_type == DataType.fitting:
            self.parent.table_loaded_from_session = True
            o_event = GeneralEventHandler(parent=self.parent)
            if o_event.is_step_selected_allowed(step_index_requested=4):
                BinningLauncher(parent=self.parent)
                FittingLauncher(parent=self.parent)
            self.parent.table_loaded_from_session = False



#!/usr/bin/env python
"""
Manual right click
"""

import logging

from qtpy import QtGui
from qtpy.QtWidgets import QMenu

from ibeatles.tools.utilities import TimeSpectraKeys
from ibeatles.utilities.table_handler import TableHandler


class ManualRightClick:
    def __init__(self, parent=None):
        self.parent = parent
        self.logger = logging.getLogger("maverick")

    def manual_table_right_click(self):
        o_table = TableHandler(table_ui=self.parent.ui.bin_manual_tableWidget)
        last_row = o_table.row_count()

        menu = QMenu(self.parent)

        # row_selected = o_table.get_row_selected()
        remove_bin = -1
        # clean_sort = None
        # load_table = menu.addAction("Import table ...")

        if last_row > 0:
            menu.addSeparator()
            remove_bin = menu.addAction("Remove selected bin")
            # clean_sort = menu.addAction("Sort and remove duplicates")

        action = menu.exec_(QtGui.QCursor.pos())
        if action == remove_bin:
            self.remove_selected_bin()
            self.parent.update_statistics()
        # elif action == load_table:
        #     self.load_manual_bin_table()
        # elif action == clean_sort:
        #     self.sort_and_remove_duplicates()

        else:
            pass

    # def load_manual_bin_table(self):
    #     o_load = LoadBinTable(parent=self.parent)
    #     o_load.run()

    def remove_selected_bin(self):
        """
        remove from the manual table the bin selected
        """
        o_table = TableHandler(table_ui=self.parent.ui.bin_manual_tableWidget)
        row_selected = o_table.get_row_selected()
        item_to_remove = self.parent.list_of_manual_bins_item[row_selected]
        self.parent.bin_profile_view.removeItem(item_to_remove)
        self.parent.list_of_manual_bins_item.pop(row_selected)
        o_table.remove_row(row=row_selected)
        self.logger.info(f"User manually removed row: {row_selected}")

        # remove the bins for statistics table
        self.parent.manual_bins[TimeSpectraKeys.file_index_array].pop(row_selected)
        self.parent.manual_bins[TimeSpectraKeys.tof_array].pop(row_selected)
        self.parent.manual_bins[TimeSpectraKeys.lambda_array].pop(row_selected)



#!/usr/bin/env python
"""
Plot module
"""

import copy

import numpy as np

from ibeatles.tools import ANGSTROMS, LAMBDA, MICRO
from ibeatles.tools.tof_bin import TO_ANGSTROMS_UNITS, TO_MICROS_UNITS
from ibeatles.tools.tof_bin.utilities.get import Get
from ibeatles.tools.utilities import TimeSpectraKeys


class Plot:
    def __init__(self, parent=None):
        self.parent = parent

    def refresh_profile_plot_and_clear_bins(self):
        """
        this clear, remove all bin items and just replot the profile on the bin right imageView
        """
        x0 = self.parent.bin_roi["x0"]
        y0 = self.parent.bin_roi["y0"]
        width = self.parent.bin_roi["width"]
        height = self.parent.bin_roi["height"]

        self.parent.bin_profile_view.clear()  # clear previous plot
        if self.parent.dict_of_bins_item is not None:  # remove previous bins
            for _key in self.parent.dict_of_bins_item.keys():
                self.parent.bin_profile_view.removeItem(self.parent.dict_of_bins_item[_key])
            self.parent.dict_of_bins_item = None

        array_of_data = self.parent.images_array

        profile_signal = [np.mean(_data[y0 : y0 + height, x0 : x0 + width]) for _data in array_of_data]
        # profile_signal = self.parent.profile_signal

        o_get = Get(parent=self.parent)
        time_spectra_x_axis_name = o_get.x_axis_selected()

        x_axis = copy.deepcopy(self.parent.time_spectra[time_spectra_x_axis_name])

        if time_spectra_x_axis_name == TimeSpectraKeys.file_index_array:
            x_axis_label = "file index"
        elif time_spectra_x_axis_name == TimeSpectraKeys.tof_array:
            x_axis *= TO_MICROS_UNITS  # to display axis in micros
            x_axis_label = "tof (" + MICRO + "s)"
        elif time_spectra_x_axis_name == TimeSpectraKeys.lambda_array:
            x_axis *= TO_ANGSTROMS_UNITS  # to display axis in Angstroms
            x_axis_label = LAMBDA + "(" + ANGSTROMS + ")"

        self.parent.bin_profile_view.plot(x_axis, profile_signal, pen="r", symbol="x")
        self.parent.bin_profile_view.setLabel("bottom", x_axis_label)
        self.parent.bin_profile_view.setLabel("left", "mean counts of ROI")



#!/usr/bin/env python
"""Pydantic configuration model for CLI and GUI application settings (user)"""

import logging
from enum import Enum
from pathlib import Path
from typing import Dict, List, Literal, Optional, Tuple, Union

import matplotlib.pyplot as plt
from pydantic import BaseModel, Field, field_validator, model_validator


class SampleBackground(BaseModel):
    x0: int
    y0: int
    width: int
    height: int


class KernelType(str, Enum):
    box = "Box"
    gaussian = "Gaussian"


class KernelSize(BaseModel):
    y: int = 3
    x: int = 3
    lambda_: int = Field(default=3, alias="lambda")

    @model_validator(mode="after")
    def check_size(cls, values):
        return values


class MovingAverage(BaseModel):
    active: bool = True
    dimension: Literal["2D", "3D"] = "2D"
    size: KernelSize = Field(default_factory=KernelSize)
    type: KernelType = KernelType.box

    @model_validator(mode="after")
    def check_size(self) -> "MovingAverage":
        if self.dimension == "2D" and hasattr(self.size, "lambda_"):
            delattr(self.size, "lambda_")
        return self


class ProcessOrder(str, Enum):
    moving_average_normalization = "Moving average, Normalization"
    normalization_moving_average = "Normalization, Moving Average"


class NormalizationConfig(BaseModel):
    sample_background: Optional[List[SampleBackground]] = None
    moving_average: MovingAverage = Field(default_factory=MovingAverage)
    processing_order: ProcessOrder = ProcessOrder.moving_average_normalization


class PixelBinning(BaseModel):
    x0: int
    y0: int
    width: int
    height: int
    bins_size: int


class BinCoordinates(BaseModel):
    """Model for bin coordinates."""

    x0: int
    x1: int
    y0: int
    y1: int
    row_index: int  # Row number in binning grid
    column_index: int  # Column number in binning grid

    @model_validator(mode="after")
    def validate_coordinates(self) -> "BinCoordinates":
        """Validate bin coordinates."""
        if self.x0 >= self.x1:
            raise ValueError("x0 must be less than x1")
        if self.y0 >= self.y1:
            raise ValueError("y0 must be less than y1")
        if self.row_index < 0 or self.column_index < 0:
            raise ValueError("Row and column indices must be non-negative")
        return self


class ThresholdFinder(BaseModel):
    method: Literal["Sliding Average", "Error Function", "Change Point"] = "Sliding Average"
    threshold_width: int = 5


class FittingCriteria(BaseModel):
    """Fit results must be smaller than the threshold to be considered valid, if 'use' is True"""

    lambda_hkl: Dict[Literal["use", "value"], Union[bool, float]] = {
        "use": True,
        "value": 0.010,
    }
    tau: Dict[Literal["use", "value"], Union[bool, float]] = {
        "use": False,
        "value": 0.010,
    }
    sigma: Dict[Literal["use", "value"], Union[bool, float]] = {
        "use": False,
        "value": 0.010,
    }


class RejectionCriteria(BaseModel):
    bragg_peak_lambda_min: Dict[Literal["use", "value"], Union[bool, float]] = {
        "use": True,
        "value": 0.000,
    }
    bragg_peak_lambda_max: Dict[Literal["use", "value"], Union[bool, float]] = {
        "use": True,
        "value": 10.000,
    }


class FittingConfig(BaseModel):
    lambda_min: float
    lambda_max: float
    threshold_finder: ThresholdFinder = Field(default_factory=ThresholdFinder)
    fitting_criteria: FittingCriteria = Field(default_factory=FittingCriteria)
    rejection_criteria: RejectionCriteria = Field(default_factory=RejectionCriteria)


class InterpolationMethod(str, Enum):
    """Supported interpolation methods for strain mapping visualization."""

    NONE = "none"
    NEAREST = "nearest"
    BILINEAR = "bilinear"
    BICUBIC = "bicubic"
    SPLINE16 = "spline16"
    SPLINE36 = "spline36"
    HANNING = "hanning"
    HAMMING = "hamming"
    HERMITE = "hermite"
    KAISER = "kaiser"
    QUADRIC = "quadric"
    CATROM = "catrom"
    GAUSSIAN = "gaussian"
    BESSEL = "bessel"
    MITCHELL = "mitchell"
    SINC = "sinc"
    LANCZOS = "lanczos"


class FigureFormat(str, Enum):
    """Supported figure formats for saving plots."""

    PNG = "png"
    PDF = "pdf"
    SVG = "svg"
    EPS = "eps"


class CsvFormat(BaseModel):
    """Configuration for CSV output format."""

    delimiter: str = ","
    include_metadata_header: bool = True
    metadata_comment_char: str = "#"
    na_rep: str = "null"


class OutputFileConfig(BaseModel):
    """Configuration for output file formats and options."""

    strain_map_format: FigureFormat = FigureFormat.PNG
    fitting_grid_format: FigureFormat = FigureFormat.PDF
    figure_dpi: int = Field(default=300, ge=72, le=1200)
    csv_format: CsvFormat = Field(default_factory=CsvFormat)

    class Config:
        use_enum_values = True


class StrainVisualization(BaseModel):
    """Configuration for strain mapping visualization."""

    interpolation_method: InterpolationMethod = InterpolationMethod.NEAREST
    colormap: str = "viridis"
    alpha: float = Field(default=0.5, ge=0.0, le=1.0, description="Transparency for overlay plots")
    display_fit_quality: bool = True

    @field_validator("colormap")
    @classmethod
    def validate_colormap(cls, v):
        """Validate that the colormap exists in matplotlib."""
        if v not in plt.colormaps():
            raise ValueError(
                f"Colormap '{v}' is not available in matplotlib. Available colormaps: {', '.join(plt.colormaps())}"
            )
        return v


class StrainMapping(BaseModel):
    """Configuration for strain mapping analysis."""

    d0: Optional[float] = None  # optional override for d0
    quality_threshold: float = Field(
        default=0.8,
        ge=0.0,
        le=1.0,
        description="R-squared threshold for accepting fit results",
    )
    visualization: StrainVisualization = Field(
        default_factory=StrainVisualization, description="Visualization settings"
    )
    output_file_config: OutputFileConfig = Field(
        default_factory=OutputFileConfig, description="Output file format settings"
    )
    save_intermediate_results: bool = False

    @model_validator(mode="after")
    def check_d0_warning(self) -> "StrainMapping":
        """Validate d0 and add warning if using non-standard value."""
        if self.d0 is not None:
            logging.warning("Using user-provided d0. This overrides the value calculated from material properties.")
        return self

    @model_validator(mode="after")
    def ensure_d0_is_positive(self) -> "StrainMapping":
        """Ensure that d0 is positive."""
        if self.d0 is not None and self.d0 <= 0:
            raise ValueError("d0 must be a positive value")
        return self


class CustomMaterial(BaseModel):
    name: str
    lattice: float
    crystal_structure: str
    hkl_lambda_pairs: Dict[Tuple[int, int, int], float]


class Material(BaseModel):
    element: Optional[str] = None
    custom_material: Optional[CustomMaterial] = None

    @model_validator(mode="after")
    def check_material_specification(self) -> "Material":
        if self.element is None and self.custom_material is None:
            raise ValueError("Either 'element' or 'custom_material' must be specified")
        if self.element is not None and self.custom_material is not None:
            raise ValueError("Only one of 'element' or 'custom_material' should be specified")
        return self


class AnalysisConfig(BaseModel):
    material: Material
    pixel_binning: PixelBinning
    fitting: FittingConfig
    strain_mapping: StrainMapping = Field(default_factory=StrainMapping)
    distance_source_detector_in_m: float = Field(default=25.0, description="Distance from source to detector in meters")
    detector_offset_in_us: float = Field(default=5000, description="Detector offset in microseconds")


class RawData(BaseModel):
    raw_data_dir: Path
    extension: str = ".tif"

    @model_validator(mode="after")
    def check_raw_data_dir(self) -> "RawData":
        if not self.raw_data_dir.exists():
            raise ValueError(f"Directory '{self.raw_data_dir}' does not exist")
        return self

    @model_validator(mode="after")
    def check_raw_data_extension(self) -> "RawData":
        # extension must be ".tif", ".tiff", ".fits"
        if self.extension not in [".tif", ".tiff", ".fits"]:
            raise ValueError("Raw data extension must be '.tif', '.tiff', or '.fits'")
        return self


class OpenBeamData(BaseModel):
    open_beam_data_dir: Path
    extension: str = ".tif"

    @model_validator(mode="after")
    def check_open_beam_data_dir(self) -> "OpenBeamData":
        if not self.open_beam_data_dir.exists():
            raise ValueError(f"Directory '{self.open_beam_data_dir}' does not exist")
        return self

    @model_validator(mode="after")
    def check_open_beam_data_extension(self) -> "OpenBeamData":
        # extension must be ".tif", ".tiff", ".fits"
        if self.extension not in [".tif", ".tiff", ".fits"]:
            raise ValueError("Open beam data extension must be '.tif', '.tiff', or '.fits'")
        return self


class IBeatlesUserConfig(BaseModel):
    raw_data: RawData
    open_beam: Optional[OpenBeamData] = None
    spectra_file_path: Optional[Path] = None
    normalization: NormalizationConfig = Field(default_factory=NormalizationConfig)
    analysis: AnalysisConfig
    output: Dict[str, Path] = Field(...)


# Example usage:
if __name__ == "__main__":
    config_dict = {
        "raw_data": {
            "raw_data_dir": "/tmp",
            "extension": ".tif",
        },
        "open_beam": {
            "open_beam_data_dir": "/tmp",
            "extension": ".tif",
        },
        "spectra_file_path": "/tmp/spectra.txt",
        "output": {
            "normalized_data_dir": "/path/to/normalized_data",
            "analysis_results_dir": "/path/to/analysis_results",
            "strain_results_dir": "/path/to/strain_results",
        },
        "normalization": {
            "sample_background": [{"x0": 0, "y0": 0, "width": 10, "height": 10}],
            "moving_average": {
                "dimension": "3D",
                "size": {"y": 3, "x": 3, "lambda": 3},
            },
        },
        "analysis": {
            # Standard material example
            "material": {"element": "Fe"},
            # Custom material example (commented out)
            # "material": {
            #     "custom_material": {
            #         "name": "Custom Alloy",
            #         "lattice": 3.52,
            #         "crystal_structure": "BCC",
            #         "hkl_lambda_pairs": {(1, 1, 0): 2.8664, (2, 0, 0): 2.0267}
            #     }
            # },
            "pixel_binning": {
                "x0": 0,
                "y0": 0,
                "width": 100,
                "height": 100,
                "bins_size": 5,
            },
            "fitting": {"lambda_min": 0.5, "lambda_max": 5.0},
            "strain_mapping": {
                "d0": 3.52,  # optional, if not provided will use material properties
                "quality_threshold": 0.8,
                "visualization": {
                    "interpolation_method": "nearest",
                    "colormap": "viridis",
                    "alpha": 0.5,
                    "display_fit_quality": True,
                },
                "output_file_config": {
                    "strain_map_format": "png",
                    "fitting_grid_format": "pdf",
                    "figure_dpi": 300,
                    "csv_format": {
                        "delimiter": ",",
                        "include_metadata_header": True,
                        "metadata_comment_char": "#",
                    },
                },
                "save_intermediate_results": False,
            },
            "distance_source_detector_in_m": 19.855,
            "detector_offset_in_us": 9600,
        },
    }

    config = IBeatlesUserConfig(**config_dict)
    print(config.model_dump_json(indent=2))



#!/usr/bin/env python
"""
Rotate export launcher
"""

import logging
import warnings

from qtpy.QtWidgets import QApplication, QDialog

from ibeatles import DataType, load_ui
from ibeatles.tools.rotate.event_handler import EventHandler as RotateEventHandler
from ibeatles.tools.utilities.reload.reload import Reload

warnings.filterwarnings("ignore")


class RotateExportLauncher(QDialog):
    def __init__(self, parent=None, top_parent=None):
        self.parent = parent
        self.top_parent = top_parent
        QDialog.__init__(self, parent=parent)
        self.ui = load_ui("ui_rotate_export.ui", baseinstance=self)

    def ok_clicked(self):
        self.parent.ui.setEnabled(False)
        o_event = RotateEventHandler(parent=self.parent, top_parent=self.top_parent)
        output_folder = o_event.select_output_folder()
        if output_folder is None:
            logging.info("User canceled file browser!")
            self.close()

        self.close()
        full_output_folder_name = o_event.rotate_data(output_folder=output_folder)
        self.parent.close()

        if self.ui.reload_checkBox.isChecked():
            o_reload = Reload(parent=self.parent, top_parent=self.top_parent)
            o_reload.run(data_type=DataType.normalized, output_folder=full_output_folder_name)

        self.parent.ui.setEnabled(True)
        QApplication.restoreOverrideCursor()



#!/usr/bin/env python
"""
FittingInitializationHandler class for handling the initialization of the fitting parameters.
"""

import numpy as np
from qtpy import QtCore
from qtpy.QtWidgets import QApplication

from ibeatles.fitting.initialization_sigma_alpha import InitializationSigmaAlpha
from ibeatles.table_dictionary.table_dictionary_handler import (
    TableDictionaryHandler,
)


class FittingInitializationHandler(object):
    all_variables_initialized = True
    advanced_mode = False
    selection_range = {
        "left_range": {
            "x_axis": [],
            "y_axis": [],
        },
        "right_range": {
            "x_axis": [],
            "y_axis": [],
        },
        "inflection": {"x": np.nan, "y": np.nan},
    }

    a1 = np.nan  # only used when using basic mode to calculate a2

    percentage_of_data_to_remove_on_side = 10.0  # %

    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent

    def make_all_active(self):
        o_table = TableDictionaryHandler(parent=self.parent, grand_parent=self.grand_parent)
        o_table.full_table_selection_tool(status=True)
        self.grand_parent.fitting_ui.update_table()
        self.grand_parent.fitting_ui.update_bragg_edge_plot()

    def run(self):
        InitializationSigmaAlpha(parent=self.parent, grand_parent=self.grand_parent)

    def finished_up_initialization(self):
        self.advanced_mode = self.parent.ui.advanced_table_checkBox.isChecked()
        if self.parent.sigma_alpha_initialized:
            QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)
            self.retrieve_parameters_and_update_table()
            self.parent.update_table()
            QApplication.restoreOverrideCursor()

    def retrieve_parameters_and_update_table(self):
        table_handler = TableDictionaryHandler(parent=self.parent, grand_parent=self.grand_parent)
        initialization_table = self.parent.initialization_table

        d_spacing = self.get_d_spacing()
        if np.isnan(d_spacing):
            self.all_variables_initialized is False
        initialization_table["d_spacing"] = d_spacing
        table_handler.fill_table_with_variable(variable_name="d_spacing", value=d_spacing, all_keys=True)

        sigma = self.get_sigma()
        if np.isnan(sigma):
            self.all_variables_initialized is False
        table_handler.fill_table_with_variable(variable_name="sigma", value=sigma, all_keys=True)

        alpha = self.get_alpha()
        if np.isnan(alpha):
            self.all_variables_initialized is False
        table_handler.fill_table_with_variable(variable_name="alpha", value=alpha, all_keys=True)

        # this function will allow to retrieve parameters that will be used by a1, a2, a5 and a6
        self.isolate_left_and_right_part_of_inflection_point()

        if self.advanced_mode:
            a2 = self.get_a2()
            if np.isnan(a2):
                self.all_variables_initialized is False
            initialization_table["a2"] = a2
            table_handler.fill_table_with_variable(variable_name="a2", value=a2, all_keys=True)

            a5 = self.get_a5()
            if np.isnan(a5):
                self.all_variables_initialized is False
            initialization_table["a5"] = a5
            table_handler.fill_table_with_variable(variable_name="a5", value=a5, all_keys=True)

            a6 = self.get_a6()
            if np.isnan(a6):
                self.all_variables_initialized is False
            initialization_table["a6"] = a6
            table_handler.fill_table_with_variable(variable_name="a6", value=a6, all_keys=True)

            a1 = self.get_a1()
            if np.isnan(a1):
                self.all_variables_initialized is False
            initialization_table["a1"] = a1
            table_handler.fill_table_with_variable(variable_name="a1", value=a1, all_keys=True)

        else:  # basic mode
            a1 = self.get_a1()
            if np.isnan(a1):
                self.all_variables_initialized is False
            initialization_table["a1"] = a1
            table_handler.fill_table_with_variable(variable_name="a1", value=a1, all_keys=True)

            a2 = self.get_a2()
            if np.isnan(a2):
                self.all_variables_initialized is False
            initialization_table["a2"] = a2
            table_handler.fill_table_with_variable(variable_name="a2", value=a2, all_keys=True)

        self.parent.initialization_table = initialization_table

    def isolate_left_and_right_part_of_inflection_point(self):
        # get array of counts selected
        [left_index, right_index] = self.grand_parent.fitting_bragg_edge_linear_selection

        # get full x_axis
        full_x_axis = self.parent.bragg_edge_data["x_axis"]

        # get full y_axis
        full_y_axis = self.parent.bragg_edge_data["y_axis"]

        # # calculate inflexion point (index) using Ed's method
        # y_axis = full_y_axis[left_index: right_index+1]
        # inflection_point_index = calculate_inflection_point(data=y_axis)
        # print(inflection_point_index)
        # print(y_axis[inflection_point_index + left_index])

        # for now inflection is only calculated by using center of selection
        inflection_point_index = int(np.mean([left_index, right_index]))
        self.selection_range["left_range"]["y_axis"] = full_y_axis[left_index:inflection_point_index]
        self.selection_range["left_range"]["x_axis"] = full_x_axis[left_index:inflection_point_index]
        self.selection_range["right_range"]["y_axis"] = full_y_axis[inflection_point_index:]
        self.selection_range["right_range"]["x_axis"] = full_x_axis[inflection_point_index:]
        self.selection_range["inflection"]["y"] = full_y_axis[inflection_point_index]
        self.selection_range["inflection"]["x"] = full_x_axis[inflection_point_index]

    def get_a1(self):
        if self.advanced_mode:
            intercept = self.a2_intercept
            a2 = self.a2
            a6 = self.a6
            return intercept + a2 * a6
        else:
            left_range = self.selection_range["left_range"]["y_axis"]
            nbr_data = len(left_range)
            nbr_data_to_remove = int((self.percentage_of_data_to_remove_on_side / 100.0) * nbr_data)
            a1 = np.mean(left_range[0:-nbr_data_to_remove])
            self.a1 = a1
            return a1

    def get_a2(self):
        if self.advanced_mode:
            x_axis = self.selection_range["left_range"]["x_axis"]
            y_axis = self.selection_range["left_range"]["y_axis"]

            nbr_data = len(x_axis)
            nbr_data_to_remove = int((self.percentage_of_data_to_remove_on_side / 100.0) * nbr_data)

            x_axis_to_use = x_axis[0:nbr_data_to_remove]
            y_axis_to_use = y_axis[0:nbr_data_to_remove]

            [slope, interception] = np.polyfit(x_axis_to_use, y_axis_to_use, 1)
            self.a2 = slope  # saving it to calculate a6
            self.a2_intercept = interception  # saving it to calculate a1
            return slope

        else:
            _mean_left_side = self.a1
            right_range = self.selection_range["right_range"]["y_axis"]
            nbr_data = len(right_range)
            nbr_data_to_remove = int((self.percentage_of_data_to_remove_on_side / 100.0) * nbr_data)
            _mean_right_side = np.mean(right_range[nbr_data_to_remove:])
            a2 = np.abs(_mean_right_side - _mean_left_side)
            return a2

    def get_a5(self):
        x_axis = self.selection_range["right_range"]["x_axis"]
        y_axis = self.selection_range["right_range"]["y_axis"]

        nbr_data = len(x_axis)
        nbr_data_to_remove = int((self.percentage_of_data_to_remove_on_side / 100.0) * nbr_data)

        x_axis_to_use = x_axis[nbr_data_to_remove:]
        y_axis_to_use = y_axis[nbr_data_to_remove:]

        [slope, interception] = np.polyfit(x_axis_to_use, y_axis_to_use, 1)
        self.a5 = slope  # saving it to calculate a6
        return slope

    def get_a6(self):
        """See docs folder for full description of formula used to get a6"""

        intensity = self.selection_range["inflection"]["y"]
        x_edge = self.selection_range["inflection"]["x"]

        a6 = x_edge - (2.0 * intensity) / (self.a5 - self.a2)
        self.a6 = a6  # saving it to caculate a1
        return a6

    def get_sigma(self):
        sigma = self.parent.initialization_table["sigma"]
        return sigma

    def get_alpha(self):
        alpha = self.parent.initialization_table["alpha"]
        return alpha

    def get_d_spacing(self):
        """
        calculates the d-spacing using the lambda range selection and using the central lambda
        2* d_spacing = lambda
        """
        print(f"self.parent.ui.lambda_min_lineEdit.text(): {self.parent.ui.lambda_min_lineEdit.text()}")
        lambda_min = float(str(self.parent.ui.lambda_min_lineEdit.text()))
        lambda_max = float(str(self.parent.ui.lambda_max_lineEdit.text()))

        average_lambda = np.mean([lambda_min, lambda_max])
        d_spacing = average_lambda / 2.0

        return d_spacing



#!/usr/bin/env python
"""
ROI (Region of Interest)
"""

import pyqtgraph as pg

from ibeatles import DEFAULT_ROI, DataType
from ibeatles.utilities.colors import pen_color

# DEFAULT_ROI = ['default', '0', '0', '20', '20', '0']
# DEFAULT_NORMALIZATION_ROI = [True, '0', '0', '20', '20', RegionType.background]


class Roi:
    def __init__(self, parent=None, data_type=DataType.sample):
        self.parent = parent

    @staticmethod
    def get_roi(roi=DEFAULT_ROI):
        """roi is formatted as DEFAULT_ROI"""
        roi = pg.ROI([roi[1], roi[2]], [roi[3], roi[4]], pen=pen_color["0"], scaleSnap=True)
        roi.addScaleHandle([1, 1], [0, 0])
        return roi

    @staticmethod
    def get_default_roi():
        roi = Roi.get_roi(roi=DEFAULT_ROI)
        return roi

    @staticmethod
    def setup_roi_id(list_roi=[DEFAULT_ROI], roi_function=None):
        list_roi_id = []
        for _roi in list_roi:
            roi_id = Roi.get_roi(roi=_roi)
            roi_id.sigRegionChanged.connect(roi_function)
            list_roi_id.append(roi_id)
        return list_roi_id



#!/usr/bin/env python
"""
This module provides a class to get data.
"""

import os
from os.path import expanduser

from ibeatles.session import SessionKeys, SessionSubKeys


class Get:
    def __init__(self, parent=None):
        self.parent = parent

    def get_material(self):
        top_tab_index = self.parent.ui.material_top_tabWidget.currentIndex()
        if top_tab_index == 0:  # pre-defined
            return self.parent.ui.pre_defined_list_of_elements.currentText()
        else:  # custom
            return self.parent.ui.user_defined_element_name.text()

    def get_log_file_name(self):
        log_file_name = self.parent.config["log_file_name"]
        full_log_file_name = Get.get_full_home_file_name(log_file_name)
        return full_log_file_name

    def get_automatic_config_file_name(self):
        config_file_name = self.parent.config["session_file_name"]
        full_config_file_name = Get.get_full_home_file_name(config_file_name)
        return full_config_file_name

    @staticmethod
    def get_full_home_file_name(base_file_name):
        home_folder = expanduser("~")
        full_log_file_name = os.path.join(home_folder, base_file_name)
        return full_log_file_name

    def distance_source_detector(self) -> str:
        session_dict = self.parent.session_dict
        return str(session_dict[SessionKeys.instrument][SessionSubKeys.distance_source_detector])

    def detector_offset(self) -> str:
        session_dict = self.parent.session_dict
        return str(session_dict[SessionKeys.instrument][SessionSubKeys.detector_value])

    def bin_size(self):
        session_dict = self.parent.session_dict
        return session_dict[SessionKeys.bin][SessionSubKeys.roi][-1]

    # def auto_bins_currently_activated(self):
    #     auto_bin_mode = self.bin_auto_mode()
    #     if auto_bin_mode == BinAutoMode.linear:
    #         return self.parent.linear_bins
    #     elif auto_bin_mode == BinAutoMode.log:
    #         return self.parent.log_bins
    #     else:
    #         raise NotImplementedError("Auto bin mode not implemented!")



import datetime

TIMESTAMP_FORMAT = "%Y-%m-%d %H:%M:%S"


def get_current_time_in_special_file_name_format():
    """format the current date and time into something like  04m_07d_2022y_08h_06mn"""
    current_time = datetime.datetime.now().strftime("%mm_%dd_%Yy_%Hh_%Mmn")
    return current_time



#!/usr/bin/env python
"""
Initialization of the step6
"""

import matplotlib
import numpy as np

matplotlib.use("Qt5Agg")
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
from qtpy.QtWidgets import QVBoxLayout

from ibeatles import ANGSTROMS, LAMBDA, SUB_0, DataType
from ibeatles.fitting import FittingTabSelected
from ibeatles.fitting.kropff import BraggPeakInitParameters
from ibeatles.session import SessionKeys, SessionSubKeys
from ibeatles.step6 import (
    CMAPS,
    DEFAULT_CMAP_INDEX,
    DEFAULT_INTERPOLATION_INDEX,
    INTERPOLATION_METHODS,
    ParametersToDisplay,
)
from ibeatles.step6.get import Get
from ibeatles.utilities.mplcanvas import MplCanvasColorbar
from ibeatles.widgets.qrangeslider import QRangeSlider


class Initialization:
    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent

    def all(self):
        self.labels()
        self.parameters()
        # self.pyqtgraph()
        self.matplotlib()
        self.data()
        self.combobox()

    def combobox(self):
        self.parent.ui.interpolation_comboBox.blockSignals(True)
        self.parent.ui.interpolation_comboBox.addItems(INTERPOLATION_METHODS)
        self.parent.ui.interpolation_comboBox.setCurrentIndex(DEFAULT_INTERPOLATION_INDEX)
        self.parent.ui.interpolation_comboBox.blockSignals(False)
        self.parent.ui.interpolation_groupBox.setVisible(True)

        self.parent.ui.cmap_comboBox.blockSignals(True)
        self.parent.ui.cmap_comboBox.addItems(CMAPS)
        self.parent.ui.cmap_comboBox.setCurrentIndex(DEFAULT_CMAP_INDEX)
        self.parent.ui.cmap_comboBox.blockSignals(False)

    def labels(self):
        self.parent.ui.range_selected_from_units_label.setText(ANGSTROMS)
        self.parent.ui.range_selected_to_units_label.setText(ANGSTROMS)
        self.parent.ui.d0_units1_label.setText(ANGSTROMS)
        self.parent.ui.d0_units2_label.setText(ANGSTROMS)
        self.parent.ui.reference_material_lambda0_units_label.setText(ANGSTROMS)
        self.parent.ui.reference_material_lambda0_label.setText(LAMBDA + SUB_0)

    def range_slider(self):
        layout = QVBoxLayout()
        self.parent.ui.range_slider = QRangeSlider(splitterWidth=100, vertical=True, min_at_the_bottom=True)

        self.parent.ui.range_slider.setMin(self.parent.slider_min)
        self.parent.ui.range_slider.setMax(self.parent.slider_max)

        o_get = Get(parent=self.parent)
        parameter_displayed = o_get.parameter_to_display()

        real_min = self.parent.min_max[parameter_displayed]["min"]
        real_max = self.parent.min_max[parameter_displayed]["max"]

        self.parent.ui.max_range_lineEdit.setText(f"{real_max:.5f}")
        self.parent.ui.min_range_lineEdit.setText(f"{real_min:.5f}")

        self.parent.ui.range_slider.setRealMin(real_min)
        self.parent.ui.range_slider.setRealMax(real_max)

        self.parent.ui.range_slider.setRealRange(real_min + 1e-10, real_max)  # trick to display full range #BUG

        self.parent.ui.range_slider.startValueChanged.connect(self.parent.range_slider_start_value_changed)
        self.parent.ui.range_slider.endValueChanged.connect(self.parent.range_slider_end_value_changed)

        layout.addWidget(self.parent.ui.range_slider)
        self.parent.ui.vertical_range_slider_widget.setLayout(layout)

        self.parent.ui.range_slider.setBackgroundStyle(
            "background: qlineargradient(x1:0, y1:0, x2:0, y2:1, stop:0 #ddd, stop:1 #333);"
        )
        self.parent.ui.range_slider.handle.setStyleSheet(
            "background: qlineargradient(x1:0, y1:0, x2:0, y2:1, stop:0 #282, stop:1 #222);"
        )

    def parameters(self):
        from_lambda = self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][
            BraggPeakInitParameters.from_lambda
        ]
        to_lambda = self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][
            BraggPeakInitParameters.to_lambda
        ]
        hkl_selected = self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][
            BraggPeakInitParameters.hkl_selected
        ]
        lambda_0 = self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][
            BraggPeakInitParameters.lambda_0
        ]
        element = self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][
            BraggPeakInitParameters.element
        ]

        self.parent.ui.from_lambda.setText(from_lambda)
        self.parent.ui.to_lambda.setText(to_lambda)
        self.parent.ui.hkl_value.setText(hkl_selected)
        self.parent.ui.d0_value.setText("{:04.4f}".format(lambda_0 / 2.0))
        self.parent.ui.material_name.setText(element)
        self.parent.ui.lambda_0.setText("{:04.4f}".format(lambda_0))
        self.parent.ui.d0_user_value.setText("{:04.4f}".format(lambda_0 / 2.0))

        self.parent.bin_size = self.grand_parent.session_dict[SessionKeys.bin][SessionSubKeys.bin_size]

    def data(self):
        live_data = self.grand_parent.data_metadata[DataType.normalized]["data"]
        integrated_image = np.mean(live_data, 0)
        self.parent.integrated_image = np.transpose(integrated_image)
        [self.parent.image_size["height"], self.parent.image_size["width"]] = np.shape(integrated_image)

    # def pyqtgraph(self):
    #     image_view = pg.ImageView(view=pg.PlotItem())
    #     image_view.ui.roiBtn.hide()
    #     image_view.ui.menuBtn.hide()
    #     self.parent.image_view = image_view

    def matplotlib(self):
        def _matplotlib(parent=None, widget=None):
            sc = MplCanvasColorbar(parent, width=5, height=2, dpi=100)
            # sc.axes.plot([0,1,2,3,4,5], [10, 1, 20 ,3, 40, 50])
            toolbar = NavigationToolbar(sc, parent)
            layout = QVBoxLayout()
            layout.addWidget(toolbar)
            layout.addWidget(sc)
            widget.setLayout(layout)
            return sc

        self.parent.matplotlib_plot = _matplotlib(parent=self.parent, widget=self.parent.ui.matplotlib_widget)

        self.parent.matplotlib_interpolation_plot = _matplotlib(
            parent=self.parent, widget=self.parent.ui.matplotlib_interpolation_widget
        )

    def min_max_values(self):
        [_, x0, y0, width, height, bin_size] = self.grand_parent.session_dict[DataType.bin][SessionSubKeys.roi]

        max_row_index = self.grand_parent.session_dict[SessionKeys.bin][SessionSubKeys.nbr_row]
        max_col_index = self.grand_parent.session_dict[SessionKeys.bin][SessionSubKeys.nbr_column]

        d_array = self.parent.d_array
        d_array_roi = d_array[y0 : y0 + (max_row_index * bin_size), x0 : x0 + (max_col_index * bin_size)]
        self.parent.min_max[ParametersToDisplay.d] = {
            "min": np.nanmin(d_array_roi),
            "max": np.nanmax(d_array_roi),
            "global_min": np.nanmin(d_array_roi),
            "global_max": np.nanmax(d_array_roi),
        }

        # self.parent.min_max[ParametersToDisplay.d] = {'min': np.min(d_array_roi),
        #                                               'max': np.max(d_array_roi),
        #                                               'global_min': np.min(d_array),
        #                                               'global_max': np.max(d_array)}

        o_get = Get(parent=self.parent)
        strain_mapping = o_get.strain_mapping()
        strain_mapping_roi = strain_mapping[y0 : y0 + (max_row_index * bin_size), x0 : x0 + (max_col_index * bin_size)]
        self.parent.min_max[ParametersToDisplay.strain_mapping] = {
            "min": np.nanmin(strain_mapping_roi),
            "max": np.nanmax(strain_mapping_roi),
            "global_min": np.nanmin(strain_mapping_roi),
            "global_max": np.nanmax(strain_mapping_roi),
        }

        # self.parent.min_max[ParametersToDisplay.strain_mapping] = {'min': np.min(strain_mapping_roi),
        #                                                            'max': np.max(strain_mapping_roi),
        #                                                            'global_min': np.min(strain_mapping),
        #                                                            'global_max': np.max(strain_mapping)}



from ibeatles.app.models.time_spectra_model import TimeSpectraModel
from ibeatles.app.ui.time_spectra_view import TimeSpectraView
from ibeatles.utilities.file_handler import FileHandler


class TimeSpectraPresenter:
    def __init__(self, parent):
        self.parent = parent
        self.model = TimeSpectraModel()
        self.view = TimeSpectraView(self)

    def load_data(
        self,
        file_path: str,
        distance_source_detector_m: float,
        detector_offset_micros: float,
    ):
        try:
            self.model.load_data(file_path, distance_source_detector_m, detector_offset_micros)
            self.update_view()
        except FileNotFoundError:
            self.view.show_error("File not found", f"The file {file_path} could not be found.")
        except Exception as e:
            self.view.show_error("Error loading data", str(e))

    def update_view(self):
        data = self.model.get_data()
        self.view.plot_data(data["tof_array"], data["counts_array"], data["lambda_array"])
        self.view.set_window_title(data["short_filename"])
        file_content = FileHandler.retrieve_ascii_contain(data["filename"])
        self.view.set_text_content(file_content)

    def show_view(self):
        self.view.show()

    def on_view_closed(self):
        self.parent.time_spectra_ui = None



#!/usr/bin/env python
"""
This module provides a class to handle errors.
"""

from ibeatles.fitting.kropff import FittingRegions


class FittingErrorException(Exception):
    def __init__(self, fitting_region=FittingRegions.high_lambda, message=""):
        self.message = message
        self.fitting_region = fitting_region
        super().__init__(self.message)

    def __str__(self):
        return f"{self.fitting_region} -> {self.message}!"


class HighLambdaFittingError(FittingErrorException):
    pass


class LowLambdaFittingError(FittingErrorException):
    pass


class BraggPeakFittingError(FittingErrorException):
    pass



#!/usr/bin/env python
"""
SaveBinTab class
"""

from loguru import logger

from ibeatles import BINNING_LINE_COLOR, DEFAULT_ROI, DataType
from ibeatles.session import SessionKeys, SessionSubKeys
from ibeatles.session.save_tab import SaveTab
from ibeatles.utilities.pyqrgraph import Pyqtgrah as PyqtgraphUtilities


class SaveBinTab(SaveTab):
    def bin(self):
        """record the ROI selected"""

        if not self.parent.data_metadata[DataType.bin]["ui_accessed"]:
            return

        def format_numpy_array_into_list(numpy_array=None):
            if numpy_array is not None:
                formatted_array = []
                for _entry in numpy_array:
                    _new_entry = [int(value) for value in _entry]
                    formatted_array.append(_new_entry)
                return list(formatted_array)
            else:
                return None

        if self.parent.binning_roi is None:
            self.parent.binning_roi = DEFAULT_ROI

        [name, x0, y0, width, height, bin_size] = self.parent.session_dict[DataType.bin][SessionSubKeys.roi]

        binning_line_view_pos = self.parent.binning_line_view["pos"]
        formatted_binning_line_view_pos = format_numpy_array_into_list(binning_line_view_pos)

        binning_line_view_adj = self.parent.binning_line_view["adj"]
        formatted_binning_line_view_adj = format_numpy_array_into_list(binning_line_view_adj)

        binning_line_view_line_color = BINNING_LINE_COLOR

        if self.parent.binning_line_view["image_view"]:
            o_pyqt = PyqtgraphUtilities(
                parent=self.parent,
                image_view=self.parent.binning_line_view["image_view"],
                data_type=DataType.bin,
            )
            state = o_pyqt.get_state()
            o_pyqt.save_histogram_level(data_type_of_data=DataType.normalized)
            histogram = self.parent.image_view_settings[DataType.bin]["histogram"]
        else:
            state = None
            histogram = None

        logger.info("Recording parameters of bin tab")
        logger.info(f" x0:{x0}, y0:{y0}, width:{width}, height:{height}, bin_size:{bin_size}")
        if binning_line_view_pos is not None:
            logger.info(f" len(binning_line_view_pos): {len(binning_line_view_pos)}")
        else:
            logger.info(" binning_line_view_pos: None")

        if formatted_binning_line_view_adj is not None:
            logger.info(f" len(binning_line_view_adj): {len(binning_line_view_adj)}")
        else:
            logger.info(" binning_line_view_adj: None")

        logger.info(f" binning_line_view_line_color: {binning_line_view_line_color}")
        logger.info(f" state: {state}")
        logger.info(f" histogram: {histogram}")

        self.session_dict[SessionKeys.bin][SessionSubKeys.roi] = [
            name,
            x0,
            y0,
            width,
            height,
            bin_size,
        ]
        self.session_dict[SessionKeys.bin][SessionSubKeys.binning_line_view]["pos"] = formatted_binning_line_view_pos
        self.session_dict[SessionKeys.bin][SessionSubKeys.binning_line_view]["adj"] = formatted_binning_line_view_adj
        self.session_dict[SessionKeys.bin][SessionSubKeys.binning_line_view]["line color"] = (
            binning_line_view_line_color
        )
        self.session_dict[DataType.bin][SessionSubKeys.image_view_state] = state
        self.session_dict[DataType.bin][SessionSubKeys.image_view_histogram] = histogram
        self.session_dict[DataType.bin][SessionSubKeys.ui_accessed] = self.parent.data_metadata[DataType.bin][
            "ui_accessed"
        ]



#!/usr/bin/env python
"""
Reduction settings handler
"""

import os

from qtpy.QtWidgets import QDialog

from ibeatles import load_ui
from ibeatles.session import ReductionDimension, ReductionType, SessionSubKeys


class ReductionSettingsHandler(QDialog):
    default_kernel_size = {"x": 3, "y": 3, "l": 3}
    default_kernel_size_label = {
        "3D": "y:{}  x:{}  \u03bb:{}".format(
            default_kernel_size["y"], default_kernel_size["x"], default_kernel_size["l"]
        ),
        "2D": f"y:{default_kernel_size['y']}  x:{default_kernel_size['x']}",
    }

    def __init__(self, parent=None):
        self.parent = parent
        QDialog.__init__(self, parent=parent)
        ui_full_path = os.path.join(os.path.dirname(__file__), os.path.join("ui", "ui_reduction_settings.ui"))
        self.ui = load_ui(ui_full_path, baseinstance=self)
        self.setWindowTitle("Load previous session?")
        self.ui.pushButton.setFocus(True)
        self.initialization()

    def initialization(self):
        reduction_dict = self.parent.session_dict["reduction"]
        self.ui.activate_moving_average_checkBox.setChecked(reduction_dict["activate"])
        self.activate_moving_average_clicked()
        if reduction_dict["dimension"] == ReductionDimension.threed:
            self.ui.kernel_dimension_3d_radioButton.setChecked(True)
        if reduction_dict["size"]["flag"] == "custom":
            self.ui.kernel_size_custom_radioButton.setChecked(True)
        self.ui.kernel_size_custom_y_spinBox.setValue(int(reduction_dict["size"]["y"]))
        self.ui.kernel_size_custom_x_spinBox.setValue(int(reduction_dict["size"]["x"]))
        self.ui.kernel_size_custom_lambda_spinBox.setValue(int(reduction_dict["size"]["l"]))
        self.size_radio_button_clicked()
        self.dimension_radio_button_clicked()
        if reduction_dict["type"] == ReductionType.gaussian:
            self.ui.kernel_type_gaussian_radioButton.setChecked(True)
        if reduction_dict[SessionSubKeys.process_order] == "option1":
            self.ui.processes_order_option1_radio_button.setChecked(True)
        else:
            self.ui.processes_order_option2_radio_button.setChecked(True)

    def activate_moving_average_clicked(self):
        state = self.ui.activate_moving_average_checkBox.isChecked()
        self.ui.dimension_groupBox.setEnabled(state)
        self.ui.size_groupBox.setEnabled(state)
        self.ui.type_groupBox.setEnabled(state)
        self.ui.processing_order_groupBox.setEnabled(state)

    def dimension_radio_button_clicked(self):
        is_3d_clicked = self.ui.kernel_dimension_3d_radioButton.isChecked()
        if is_3d_clicked:
            kernel_size = ReductionDimension.threed
        else:
            kernel_size = ReductionDimension.twod
        self.ui.kernel_size_default_label.setText(self.default_kernel_size_label[kernel_size])
        self.ui.kernel_size_custom_lambda_label.setVisible(is_3d_clicked)
        self.ui.kernel_size_custom_lambda_spinBox.setVisible(is_3d_clicked)

    def size_radio_button_clicked(self):
        is_default_clicked = self.ui.kernel_size_default_radioButton.isChecked()
        self.ui.kernel_size_default_label.setEnabled(is_default_clicked)
        self.ui.kernel_size_custom_y_label.setEnabled(not is_default_clicked)
        self.ui.kernel_size_custom_y_spinBox.setEnabled(not is_default_clicked)
        self.ui.kernel_size_custom_x_label.setEnabled(not is_default_clicked)
        self.ui.kernel_size_custom_x_spinBox.setEnabled(not is_default_clicked)
        self.ui.kernel_size_custom_lambda_label.setEnabled(not is_default_clicked)
        self.ui.kernel_size_custom_lambda_spinBox.setEnabled(not is_default_clicked)

    def ok_clicked(self):
        reduction_dict = self.parent.session_dict["reduction"]
        reduction_dict["activate"] = self.ui.activate_moving_average_checkBox.isChecked()
        reduction_dict["dimension"] = (
            ReductionDimension.twod
            if self.ui.kernel_dimension_2d_radioButton.isChecked()
            else ReductionDimension.threed
        )
        reduction_dict["size"]["flag"] = "default" if self.ui.kernel_size_default_radioButton.isChecked() else "custom"
        reduction_dict["size"]["y"] = self.ui.kernel_size_custom_y_spinBox.value()
        reduction_dict["size"]["x"] = self.ui.kernel_size_custom_x_spinBox.value()
        reduction_dict["size"]["l"] = self.ui.kernel_size_custom_lambda_spinBox.value()
        reduction_dict["type"] = (
            ReductionType.box if self.ui.kernel_type_box_radioButton.isChecked() else ReductionType.gaussian
        )
        reduction_dict[SessionSubKeys.process_order] = (
            "option1" if self.ui.processes_order_option1_radio_button.isChecked() else "option2"
        )

        self.parent.session_dict["reduction"] = reduction_dict

        self.close()



#!/usr/bin/env python
"""
CreateFittingStoryLauncher class for launching the fitting story window.
"""

import numpy as np
from qtpy import QtGui
from qtpy.QtWidgets import (
    QCheckBox,
    QHBoxLayout,
    QMainWindow,
    QProgressBar,
    QTableWidgetItem,
    QTableWidgetSelectionRange,
    QWidget,
)

from ibeatles import down_image, load_ui, up_image
from ibeatles.fitting.fitting_job_handler import FittingJobHandler
from ibeatles.table_dictionary.table_fitting_story_dictionary_handler import (
    TableFittingStoryDictionaryHandler,
)


class CreateFittingStoryLauncher(object):
    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent

        if self.grand_parent.fitting_story_ui is None:
            fitting_story_window = FittingStoryWindow(parent=parent, grand_parent=grand_parent)
            fitting_story_window.show()
            self.grand_parent.fitting_story_ui = fitting_story_window

        else:
            self.grand_parent.fitting_story_ui.setFocus()
            self.grand_parent.fitting_story_ui.activateWindow()


class FittingStoryWindow(QMainWindow):
    list_column_tag = [
        "d_spacing",
        "sigma",
        "alpha",
        "a1",
        "a2",
        "a5",
        "a6",
    ]

    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent
        QMainWindow.__init__(self, parent=grand_parent)
        self.ui = load_ui("ui_fittingStoryTable.ui", baseinstance=self)
        # self.ui= UiMainWindow()
        # self.ui.setupUi(self)

        self.init_widgets()
        self.init_statusbar()
        self.initialize_table()
        self.fill_table()
        self.select_row(row=0)
        self.check_status_buttons(row=0)

    def init_statusbar(self):
        self.eventProgress2 = QProgressBar(self.ui.statusbar)
        self.eventProgress2.setMinimumSize(20, 14)
        self.eventProgress2.setMaximumSize(540, 100)
        self.eventProgress2.setVisible(False)
        self.ui.statusbar.addPermanentWidget(self.eventProgress2)

        self.eventProgress = QProgressBar(self.ui.statusbar)
        self.eventProgress.setMinimumSize(20, 14)
        self.eventProgress.setMaximumSize(540, 100)
        self.eventProgress.setVisible(False)
        self.ui.statusbar.addPermanentWidget(self.eventProgress)

    def init_widgets(self):
        icon_up = QtGui.QIcon(up_image)
        self.ui.up_button.setIcon(icon_up)
        self.ui.up_button.setEnabled(False)

        icon_down = QtGui.QIcon(down_image)
        self.ui.down_button.setIcon(icon_down)
        self.ui.down_button.setEnabled(False)

    def clear_table(self):
        nbr_row = self.ui.story_table.rowCount()
        for _row in np.arange(nbr_row):
            self.ui.story_table.removeRow(0)

    def initialize_table(self):
        if self.grand_parent.table_fitting_story_dictionary == {}:
            o_table = TableFittingStoryDictionaryHandler(parent=self.parent, grand_parent=self.grand_parent)
            o_table.initialize_table()

    def reset_table(self):
        o_table = TableFittingStoryDictionaryHandler(parent=self.parent, grand_parent=self.grand_parent)
        o_table.initialize_table()
        self.fill_table()
        self.select_row(row=0)
        self.check_status_buttons(row=0)

    def fill_table(self):
        table_fitting_story_dictionary = self.grand_parent.table_fitting_story_dictionary

        story_table = self.ui.story_table

        if self.grand_parent.fitting_ui.ui.advanced_table_checkBox.isChecked():
            story_table.showColumn(5)
            story_table.showColumn(6)
        else:
            story_table.hideColumn(5)
            story_table.hideColumn(6)

        self.clear_table()
        for _index in table_fitting_story_dictionary.keys():
            story_table.insertRow(_index)

            _entry = table_fitting_story_dictionary[_index]

            # _status_of_row = _entry['progress']
            # if _status_of_row == Status.in_progress:
            #     _color = QtGui.QColor(0,255,0, alpha=255)
            # elif _status_of_row == Status.completed:
            #     _color = QtGui.QColor(0, 0, 255, alpha=255)
            # else:
            #     _color = QtGui.QColor(0, 0, 0)

            # _item = self.set_item(text=str(_index+1))
            # _item.setTextColor(_color)
            # story_table.setItem(_index, 0, _item)

            for _index_tag, _tag in enumerate(self.list_column_tag):
                _widget = self.set_widget(status=_entry[_tag], row=_index, column=_index_tag)
                story_table.setCellWidget(_index, _index_tag, _widget)

    def set_item(self, text=""):
        _item = QTableWidgetItem(text)
        return _item

    def widget_state_changed(self, state=0, row=0, column=0):
        table_fitting_story_dictionary = self.grand_parent.table_fitting_story_dictionary
        _entry = table_fitting_story_dictionary[row]

        # _widget = self.ui.story_table.cellWidget(row, column)
        # list_children = _widget.children()

        status = False
        if state == 2:
            status = True

        _entry[self.list_column_tag[column]] = status
        table_fitting_story_dictionary[row] = _entry
        self.grand_parent.table_fitting_story_dictionary = table_fitting_story_dictionary

    def set_widget(self, status=False, row=0, column=0):
        _layout = QHBoxLayout()
        _widget = QCheckBox()
        _widget.stateChanged.connect(
            lambda state=0, row=row, column=column: self.widget_state_changed(state=state, row=row, column=column)
        )
        _widget.blockSignals(True)
        _widget.setChecked(status)
        _widget.blockSignals(False)
        _layout.addStretch()
        _layout.addWidget(_widget)
        _layout.addStretch()
        _new_widget = QWidget()
        _new_widget.setLayout(_layout)
        return _new_widget

    def add_row_button_clicked(self):
        list_selection = self.ui.story_table.selectedRanges()
        if len(list_selection) > 0:
            selection = self.ui.story_table.selectedRanges()[0]
            row = selection.topRow()
        else:
            row = -1

        o_table_handler = TableFittingStoryDictionaryHandler(parent=self.parent, grand_parent=self.grand_parent)
        o_table_handler.add_entry(index_to_add=row + 1)

        self.fill_table()

        self.select_row(row=row + 1)  # new row is added below selection
        self.check_status_add_remove_buttons()

    def remove_row_button_clicked(self):
        selection = self.ui.story_table.selectedRanges()[0]
        row = selection.topRow()

        o_table_handler = TableFittingStoryDictionaryHandler(parent=self.parent, grand_parent=self.grand_parent)
        o_table_handler.remove_entry(index_to_remove=row)

        self.fill_table()

        # select new row
        nbr_row = self.ui.story_table.rowCount()
        if nbr_row == 0:
            self.check_status_buttons()
            return

        elif row == nbr_row:
            new_row_selected = row - 1
        else:
            new_row_selected = row

        self.select_row(row=new_row_selected)
        self.check_status_add_remove_buttons()

    def start_fitting_button_clicked(self):
        o_fitting = FittingJobHandler(parent=self.parent, grand_parent=self.grand_parent)
        o_fitting.run_story()

    def _move_row_clicked(self, direction="up"):
        selection = self.ui.story_table.selectedRanges()[0]
        row = selection.topRow()

        o_table_handler = TableFittingStoryDictionaryHandler(parent=self.parent, grand_parent=self.grand_parent)
        o_table_handler.move_entry(current_index_row=row, direction=direction)

        self.fill_table()

        # make selection follows new position of row
        if direction == "up":
            new_row = row - 1
        else:
            new_row = row + 1
        self.select_row(row=new_row)

        self.check_status_buttons(row=new_row)

    def deselect_row(self, row=0):
        self.select_row_status(row=row, status=False)

    def select_row(self, row=0):
        self.select_row_status(row=row, status=True)

    def select_row_status(self, row=0, status=False):
        nbr_column = self.ui.story_table.columnCount()
        _selection = QTableWidgetSelectionRange(row, 0, row, nbr_column - 1)
        self.ui.story_table.setRangeSelected(_selection, status)

    def move_row_up_clicked(self):
        self._move_row_clicked(direction="up")

    def move_row_down_clicked(self):
        self._move_row_clicked(direction="down")

    def cell_clicked(self, row, column):
        self.check_status_buttons(row=row)

    def check_status_buttons(self, row=np.nan):
        """
        check the enabled status of the arrow buttons according to row clicked
        """
        nbr_row = self.ui.story_table.rowCount()

        up_status = True
        down_status = True
        start_fitting_status = True

        if nbr_row == 0:
            down_status = False
            up_status = False
            start_fitting_status = False

        elif row == (nbr_row - 1):  # we clicked the last row, disable move down
            down_status = False

        elif row == 0:  # we can't move up this row
            up_status = False

        elif np.isnan(row):
            up_status = False
            down_status = False

        # if self.parent.table_fitting_story_dictionary == {}:
        # up_status= False
        # down_status = False
        # start_fitting_status = False

        self.ui.down_button.setEnabled(down_status)
        self.ui.up_button.setEnabled(up_status)
        self.ui.start_fits_button.setEnabled(start_fitting_status)

        self.check_status_add_remove_buttons()

    def check_status_add_remove_buttons(self):
        nbr_row = self.ui.story_table.rowCount()

        add_status = True
        remove_status = True

        if nbr_row == 0:
            remove_status = False

        self.ui.add_row_button.setEnabled(add_status)
        self.ui.remove_row_button.setEnabled(remove_status)

    def closeEvent(self, event=None):
        self.grand_parent.fitting_story_ui = None



#!/usr/bin/env python
"""
Load files module
"""

from qtpy.QtWidgets import QApplication

from ibeatles.tools.tof_combine import SessionKeys as TofSessionKeys
from ibeatles.tools.tof_combine.utilities.get import Get
from ibeatles.tools.tof_combine.utilities.image_handler import ImageHandler


class LoadFiles:
    def __init__(self, parent=None, folder=None):
        self.parent = parent
        self.folder = folder

    def retrieve_data(self):
        o_get = Get(parent=self.parent)
        _row = o_get.row_of_that_folder(folder=self.folder)
        list_of_files = self.parent.dict_data_folders[_row][TofSessionKeys.list_files]

        self.parent.eventProgress.setMinimum(0)
        self.parent.eventProgress.setMaximum(len(list_of_files))
        self.parent.eventProgress.setValue(0)
        self.parent.eventProgress.setVisible(True)

        image_array = []
        for _index, _file in enumerate(list_of_files):
            try:
                o_handler = ImageHandler(parent=self.parent, filename=_file)
                _data = o_handler.get_data()
                image_array.append(_data)
                self.parent.eventProgress.setValue(_index + 1)
                QApplication.processEvents()
            except ValueError:
                # skip this file, it's a .txt
                pass

        self.parent.eventProgress.setVisible(False)

        return image_array



#!/usr/bin/env python
"""
Display class
"""

from ibeatles.utilities.display import Display as UtilitiesDisplay


class Display:
    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent

    def display_lambda_0(self):
        pyqtgraph_ui = self.parent.bragg_edge_plot
        item = self.parent.lambda_0_item_in_bragg_edge_plot
        lambda_position = float(str(self.parent.ui.bragg_edge_calculated.text()))

        o_utility_display = UtilitiesDisplay(ui=pyqtgraph_ui)
        new_item = o_utility_display.vertical_line(item=item, x_position=lambda_position)
        self.parent.lambda_0_item_in_bragg_edge_plot = new_item



import numpy as np


class CheckingFittingConditions:
    def __init__(self, fit_conditions=None):
        self.max_l_hkl_error_value = fit_conditions["l_hkl_error"]["value"]
        self.l_hkl_error_state = fit_conditions["l_hkl_error"]["state"]

        self.max_t_error_value = fit_conditions["t_error"]["value"]
        self.t_error_state = fit_conditions["t_error"]["state"]

        self.max_sigma_error_value = fit_conditions["sigma_error"]["value"]
        self.sigma_error_state = fit_conditions["sigma_error"]["state"]

    def is_fitting_ok(self, l_hkl_error=None, t_error=None, sigma_error=None):
        if self.l_hkl_error_state:
            if l_hkl_error is None:
                return False

            if not np.isfinite(l_hkl_error):
                return False

            if l_hkl_error > self.max_l_hkl_error_value:
                return False

        if self.t_error_state:
            if t_error is None:
                return False

            if not np.isfinite(t_error):
                return False

            if t_error > self.max_t_error_value:
                return False

        if self.sigma_error_state:
            if sigma_error is None:
                return False

            if not np.isfinite(sigma_error):
                return False

            if sigma_error > self.max_sigma_error_value:
                return False

        return True



#!/usr/bin/env python
"""
Initialization
"""

import matplotlib
import numpy as np
import pyqtgraph as pg
from loguru import logger
from qtpy import QtCore
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import (
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QProgressBar,
    QSizePolicy,
    QSlider,
    QSpacerItem,
    QVBoxLayout,
    QWidget,
)

matplotlib.use("Qt5Agg")
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar

from ibeatles import DataType, interact_me_style, settings_image
from ibeatles.fitting import FittingTabSelected, KropffTabSelected
from ibeatles.fitting.kropff import BraggPeakInitParameters, KropffThresholdFinder
from ibeatles.fitting.kropff import SessionSubKeys as KropffSessionSubKeys
from ibeatles.utilities.get import Get
from ibeatles.utilities.mplcanvas import MplCanvas
from ibeatles.utilities.table_handler import TableHandler


class Initialization:
    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent

    def run_all(self):
        self.pyqtgraph()
        self.data_format()
        self.table_behavior()
        self.table_headers()
        self.labels()
        self.widgets()
        self.splitter()
        self.ui()
        self.global_data()
        self.statusbar()
        self.matplotlib()
        self.tab()

    def data_format(self):
        # force all the bragg edges array to be float instead of string
        bragg_edges_array = self.grand_parent.selected_element_bragg_edges_array
        self.grand_parent.selected_element_bragg_edges_array = [float(_value) for _value in bragg_edges_array]

    def splitter(self):
        # march dollase vertical splitter
        self.parent.ui.splitter_2.setStyleSheet("""
                                     QSplitter::handle{
                                     image: url(":/MPL Toolbar/vertical_splitter_handle.png");
                                     }
                                     """)
        self.parent.ui.splitter_2.setHandleWidth(15)

        # kropff
        self.parent.ui.kropff_top_horizontal_splitter.setStyleSheet("""
                                     QSplitter::handle{
                                     image: url(":/MPL Toolbar/horizontal_splitter_handle.png");
                                     }
                                     """)

        self.parent.ui.splitter.setStyleSheet("""
                                     QSplitter::handle{
                                     image: url(":/MPL Toolbar/horizontal_splitter_handle.png");
                                     }
                                     """)
        self.parent.ui.splitter.setHandleWidth(15)

        self.parent.ui.splitter_4.setSizes([500, 500])
        self.parent.ui.splitter_4.setStyleSheet("""
                                     QSplitter::handle{
                                     image: url(":/MPL Toolbar/vertical_splitter_handle.png");
                                     }
                                     """)
        self.parent.ui.splitter_4.setHandleWidth(15)

        # self.parent.ui.area.setStyleSheet("""
        #                              QSplitter::handle{
        #                              image: url(":/MPL Toolbar/horizontal_splitter_handle.png");
        #                              }
        #                              """)

        self.parent.ui.splitter_3.setStyleSheet("""
                                     QSplitter::handle{
                                     image: url(":/MPL Toolbar/horizontal_splitter_handle.png");
                                     }
                                     """)
        self.parent.ui.splitter_3.setHandleWidth(15)

    def tab(self):
        self.parent.ui.tabWidget.setCurrentIndex(1)  # show by default Kropff
        self.parent.ui.tabWidget.setTabEnabled(0, False)  # disable March-Dollase

    def statusbar(self):
        self.parent.eventProgress = QProgressBar(self.parent.ui.statusbar)
        self.parent.eventProgress.setMinimumSize(20, 14)
        self.parent.eventProgress.setMaximumSize(540, 100)
        self.parent.eventProgress.setVisible(False)
        self.parent.ui.statusbar.addPermanentWidget(self.parent.eventProgress)
        self.parent.setStyleSheet("QStatusBar{padding-left:8px;color:red;font-weight:bold;}")

    def global_data(self):
        x_axis = self.grand_parent.normalized_lambda_bragg_edge_x_axis
        self.parent.bragg_edge_data["x_axis"] = x_axis

        # self.parent.kropff_automatic_threshold_finder_algorithm = \
        #     self.grand_parent.kropff_automatic_threshold_finder_algorithm

    def table_headers(self):
        o_kropff_high_tof = TableHandler(table_ui=self.parent.ui.high_lda_tableWidget)
        column_names = [
            "row #",
            "column #",
            "a\u2080",
            "b\u2080",
            "a\u2080_error",
            "b\u2080_error",
        ]

        column_sizes = [80, 80, 100, 100, 100, 100]
        for _col_index, _col_name in enumerate(column_names):
            o_kropff_high_tof.insert_column(_col_index)
        o_kropff_high_tof.set_column_names(column_names=column_names)
        o_kropff_high_tof.set_column_sizes(column_sizes=column_sizes)

        o_kropff_low_tof = TableHandler(table_ui=self.parent.ui.low_lda_tableWidget)
        column_names = [
            "row #",
            "column #",
            "a\u2095\u2096\u2097",
            "b\u2095\u2096\u2097",
            "a\u2095\u2096\u2097_error",
            "b\u2095\u2096\u2097_error",
        ]
        column_sizes = [80, 80, 100, 100, 100, 100]
        for _col_index, _col_name in enumerate(column_names):
            o_kropff_low_tof.insert_column(_col_index)
        o_kropff_low_tof.set_column_names(column_names=column_names)
        o_kropff_low_tof.set_column_sizes(column_sizes=column_sizes)

        o_kropff_bragg_edge = TableHandler(table_ui=self.parent.ui.bragg_edge_tableWidget)
        column_names = [
            "row #",
            "column #",
            "\u03bb\u2095\u2096\u2097",
            "\u03c4",
            "\u03c3",
            "\u03bb\u2095\u2096\u2097_error",
            "\u03c4_error",
            "\u03c3_error",
        ]
        column_sizes = [80, 80, 100, 100, 100, 100, 100, 100]
        for _col_index, _col_name in enumerate(column_names):
            o_kropff_bragg_edge.insert_column(_col_index)
        o_kropff_bragg_edge.set_column_names(column_names=column_names)
        o_kropff_bragg_edge.set_column_sizes(column_sizes=column_sizes)

        # kropff table summary
        o_kropff_summary = TableHandler(table_ui=self.parent.ui.kropff_summary_tableWidget)
        for _col in np.arange(3):
            o_kropff_summary.insert_empty_column(column=0)

        kropff_column_names = [
            "",
            "\u03bb\u2095\u2096\u2097",
            "\u03bb\u2095\u2096\u2097_error",
        ]
        kropff_column_sizes = [400, 200]
        for _col_index, _col_name in enumerate(kropff_column_names):
            o_kropff_summary.insert_column(_col_index)
        o_kropff_summary.set_column_names(kropff_column_names)
        o_kropff_summary.set_column_sizes(kropff_column_sizes)

        kropff_row_names = [
            "mean",
            "median",
            "std",
            "% of cells with a value",
            "% of cells locked (error within constraints range)",
        ]

        for _row_index, _row_name in enumerate(kropff_row_names):
            o_kropff_summary.insert_empty_row(row=_row_index)
            o_kropff_summary.insert_item(row=_row_index, column=0, value=_row_name)

    def table_behavior(self):
        for _column, _width in enumerate(self.parent.header_table_columns_width):
            self.parent.ui.header_table.setColumnWidth(_column, _width)

        for _column, _width in enumerate(self.parent.fitting_table_columns_width):
            self.parent.ui.value_table.setColumnWidth(_column, _width)

        self.parent.hori_header_table = self.parent.ui.header_table.horizontalHeader()
        self.parent.hori_value_table = self.parent.ui.value_table.horizontalHeader()

        self.parent.hori_header_table.sectionResized.connect(self.parent.resizing_header_table)
        self.parent.hori_value_table.sectionResized.connect(self.parent.resizing_value_table)

        self.parent.hori_header_table.sectionClicked.connect(self.parent.column_header_table_clicked)
        self.parent.hori_value_table.sectionClicked.connect(self.parent.column_value_table_clicked)

    def pyqtgraph(self):
        # Kropff

        # if (
        #     len(self.grand_parent.data_metadata["normalized"]["data_live_selection"])
        #     > 0
        # ) and self.grand_parent.binning_line_view["pos"] is not None:
        #     status = True
        # else:
        #     status = False

        # area = DockArea()
        # self.parent.ui.area = area
        # area.setVisible(status)
        # d1 = Dock("Image Preview", size=(200, 300))
        # d2 = Dock("Bragg Edge", size=(200, 100))

        # area.addDock(d1, 'left')
        # area.addDock(d2, 'right')

        preview_widget = pg.GraphicsLayoutWidget()
        pg.setConfigOptions(antialias=True)  # this improves display

        # LEFT WIDGET
        vertical_layout = QVBoxLayout()
        preview_widget.setLayout(vertical_layout)

        # image view (top plot)
        image_view = pg.ImageView(view=pg.PlotItem())
        image_view.ui.roiBtn.hide()
        image_view.ui.menuBtn.hide()

        self.parent.image_view_vline = pg.InfiniteLine(angle=90, movable=False)
        self.parent.image_view_hline = pg.InfiniteLine(angle=0, movable=False)
        image_view.addItem(self.parent.image_view_vline)
        image_view.addItem(self.parent.image_view_hline)

        self.parent.image_view = image_view
        self.parent.image_view_item = image_view.getImageItem()
        self.parent.image_view_scene = self.parent.image_view_item.scene()
        self.parent.image_view_proxy = pg.SignalProxy(
            self.parent.image_view_scene.sigMouseMoved,
            rateLimit=60,
            slot=self.parent.mouse_moved_in_top_left_image_view,
        )

        self.grand_parent.fitting_image_view = image_view
        image_view.scene.sigMouseMoved.connect(self.parent.mouse_moved_in_image_view)
        self.parent.image_view_scene.sigMouseClicked.connect(self.parent.mouse_clicked_in_top_left_image_view)

        # left_widget = QWidget()
        vertical = QVBoxLayout()
        vertical.addWidget(image_view)

        # cursor infos
        group_xy = QGroupBox()
        group_xy.setFixedHeight(60)
        group_xy.setTitle("Cursor")

        pos_x_label = QLabel("x:")
        pos_x_label.setFixedWidth(50)
        pos_x_label.setAlignment(QtCore.Qt.AlignRight)
        pos_x_value = QLabel("N/A")
        pos_x_value.setFixedWidth(50)
        pos_x_value.setAlignment(QtCore.Qt.AlignLeft)
        self.parent.ui.kropff_pos_x_value = pos_x_value

        pos_y_label = QLabel("y:")
        pos_y_label.setFixedWidth(50)
        pos_y_label.setAlignment(QtCore.Qt.AlignRight)
        pos_y_value = QLabel("N/A")
        pos_y_value.setFixedWidth(50)
        pos_y_value.setAlignment(QtCore.Qt.AlignLeft)
        self.parent.ui.kropff_pos_y_value = pos_y_value

        hori_layout = QHBoxLayout()
        hori_layout.addWidget(pos_x_label)
        hori_layout.addWidget(pos_x_value)
        hori_layout.addWidget(pos_y_label)
        hori_layout.addWidget(pos_y_value)
        group_xy.setLayout(hori_layout)

        # bin infos
        group_bin = QGroupBox()
        group_bin.setFixedHeight(60)
        group_bin.setTitle("Bin")

        bin_x_label = QLabel("x:")
        bin_x_label.setFixedWidth(50)
        bin_x_label.setAlignment(QtCore.Qt.AlignRight)
        bin_x_value = QLabel("N/A")
        bin_x_value.setFixedWidth(50)
        bin_x_value.setAlignment(QtCore.Qt.AlignLeft)
        self.parent.ui.kropff_bin_x_value = bin_x_value

        bin_y_label = QLabel("y:")
        bin_y_label.setFixedWidth(50)
        bin_y_label.setAlignment(QtCore.Qt.AlignRight)
        bin_y_value = QLabel("N/A")
        bin_y_value.setFixedWidth(50)
        bin_y_value.setAlignment(QtCore.Qt.AlignLeft)
        self.parent.ui.kropff_bin_y_value = bin_y_value

        bin_nbr_label = QLabel("#:")
        bin_nbr_label.setFixedWidth(50)
        bin_nbr_label.setAlignment(QtCore.Qt.AlignRight)
        bin_nbr_value = QLabel("N/A")
        bin_nbr_value.setFixedWidth(50)
        bin_nbr_value.setAlignment(QtCore.Qt.AlignLeft)
        self.parent.ui.kropff_bin_nbr_value = bin_nbr_value

        hori_layout = QHBoxLayout()
        hori_layout.addWidget(bin_x_label)
        hori_layout.addWidget(bin_x_value)
        hori_layout.addWidget(bin_y_label)
        hori_layout.addWidget(bin_y_value)
        hori_layout.addWidget(bin_nbr_label)
        hori_layout.addWidget(bin_nbr_value)
        group_bin.setLayout(hori_layout)

        groups_xy_bin = QWidget()
        groups_xh_bin_layout = QHBoxLayout()
        groups_xh_bin_layout.addWidget(group_xy)
        groups_xh_bin_layout.addWidget(group_bin)
        groups_xy_bin.setLayout(groups_xh_bin_layout)

        # bin transparency
        transparency_layout = QHBoxLayout()
        # spacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        # transparency_layout.addItem(spacer)
        label = QLabel("Bin Transparency")
        transparency_layout.addWidget(label)
        slider = QSlider(QtCore.Qt.Horizontal)
        slider.setMaximum(100)
        slider.setMinimum(0)
        slider.setValue(50)
        slider.valueChanged.connect(self.parent.slider_changed)
        self.parent.slider = slider
        transparency_layout.addWidget(slider)
        transparency_widget = QWidget()
        transparency_widget.setLayout(transparency_layout)

        vertical.addWidget(groups_xy_bin)
        vertical.addWidget(transparency_widget)

        # bottom_vertical_layout = QVBoxLayout()
        # bottom_vertical_layout.addWidget(groups_xy_bin)
        # bottom_vertical_layout.addWidget(transparency_widget)
        # bottom_widget.setLayout(bottom_vertical_layout)
        # left_widget.setLayout(vertical)

        self.parent.ui.widget_kropff_left.setLayout(vertical)

        # d1.addWidget(top_widget)
        # d1.addWidget(bottom_widget)

        # RIGHT WIDGET
        bragg_edge_plot = pg.PlotWidget(title="")
        bragg_edge_plot.plot()
        self.parent.bragg_edge_plot = bragg_edge_plot

        # plot all or individual bins
        buttons_layout = QHBoxLayout()
        spacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        buttons_layout.addItem(spacer)
        label = QLabel("Plot")
        buttons_layout.addWidget(label)
        self.parent.ui.plot_label = label

        # d2.addWidget(bragg_edge_plot)
        right_vertical_layout = QVBoxLayout()
        right_vertical_layout.addWidget(bragg_edge_plot)
        self.parent.ui.widget_kropff_right.setLayout(right_vertical_layout)

        # vertical_layout.addWidget(area)
        # self.parent.ui.widget_kropff.setLayout(vertical_layout)

        # bottom right plot
        self.parent.ui.kropff_fitting = pg.PlotWidget(title="Fitting")
        fitting_layout = QVBoxLayout()
        fitting_layout.addWidget(self.parent.ui.kropff_fitting)
        self.parent.ui.kropff_widget.setLayout(fitting_layout)

    def labels(self):
        self.parent.ui.lambda_min_label.setText("\u03bb<sub>min</sub>")
        self.parent.ui.lambda_max_label.setText("\u03bb<sub>max</sub>")
        self.parent.ui.lambda_min_units.setText("\u212b")
        self.parent.ui.lambda_max_units.setText("\u212b")
        self.parent.ui.bragg_edge_units.setText("\u212b")
        self.parent.ui.bragg_edge_infos_lambda_0_label.setText("\u03bb<sub>0</sub>")

        # material name
        o_get = Get(parent=self.grand_parent)
        material_name = o_get.get_material()
        self.parent.ui.material_groupBox.setTitle(material_name)

    def widgets(self):
        """
        such as material h,k,l list according to material selected in normalized tab
        """
        kropff_session_dict = self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff]

        def _retrieve_kropff_init_value(
            variable_key: KropffSessionSubKeys = KropffSessionSubKeys.lambda_hkl,
            key: BraggPeakInitParameters = BraggPeakInitParameters.fix_flag,
        ):
            """
            return the parameter in the kropff_session_dict[bragg_peak][variable_key][key]
            """
            if key is None:
                return ""

            if not kropff_session_dict.get(KropffSessionSubKeys.bragg_peak, None):
                return None

            if not kropff_session_dict[KropffSessionSubKeys.bragg_peak].get(variable_key, None):
                return None

            return kropff_session_dict[KropffSessionSubKeys.bragg_peak][variable_key].get(key, None)

        hkl_list = self.grand_parent.selected_element_hkl_array

        str_hkl_list = ["{},{},{}".format(_hkl[0], _hkl[1], _hkl[2]) for _hkl in hkl_list]
        self.parent.ui.hkl_list_ui.addItems(str_hkl_list)

        # Kropff - initial fitting parameters
        a0 = kropff_session_dict[KropffSessionSubKeys.high_tof][KropffSessionSubKeys.a0]
        b0 = kropff_session_dict[KropffSessionSubKeys.high_tof][KropffSessionSubKeys.b0]
        high_tof_graph = kropff_session_dict[KropffSessionSubKeys.high_tof][KropffSessionSubKeys.graph]
        self.parent.ui.kropff_high_lda_a0_init.setText(a0)
        self.parent.ui.kropff_high_lda_b0_init.setText(b0)
        if high_tof_graph == "a0":
            self.parent.ui.kropff_a0_radioButton.setChecked(True)
        else:
            self.parent.ui.kropff_b0_radioButton.setChecked(True)

        ahkl = kropff_session_dict[KropffSessionSubKeys.low_tof][KropffSessionSubKeys.ahkl]
        bhkl = kropff_session_dict[KropffSessionSubKeys.low_tof][KropffSessionSubKeys.bhkl]
        low_tof_graph = kropff_session_dict[KropffSessionSubKeys.low_tof][KropffSessionSubKeys.graph]
        self.parent.ui.kropff_low_lda_ahkl_init.setText(ahkl)
        self.parent.ui.kropff_low_lda_bhkl_init.setText(bhkl)
        if low_tof_graph == "ahkl":
            self.parent.ui.kropff_ahkl_radioButton.setChecked(True)
        else:
            self.parent.ui.kropff_bhkl_radioButton.setChecked(True)

        # lambda hkl
        lambda_hkl_fix_flag = _retrieve_kropff_init_value(
            variable_key=KropffSessionSubKeys.lambda_hkl,
            key=BraggPeakInitParameters.fix_flag,
        )
        if lambda_hkl_fix_flag:
            self.parent.ui.lambda_hkl_fix_radioButton.setChecked(True)
        else:
            self.parent.ui.lambda_hkl_range_radioButton.setChecked(True)

        self.parent.kropff_initial_guess_lambda_hkl_fix_clicked()

        lambda_hkl_fix_value = _retrieve_kropff_init_value(
            variable_key=KropffSessionSubKeys.lambda_hkl,
            key=BraggPeakInitParameters.fix_value,
        )
        self.parent.ui.lambda_hkl_fix_lineEdit.setText(str(lambda_hkl_fix_value))

        lambda_hkl_range_from = _retrieve_kropff_init_value(
            variable_key=KropffSessionSubKeys.lambda_hkl,
            key=BraggPeakInitParameters.range_from,
        )
        self.parent.ui.lambda_hkl_from_lineEdit.setText(str(lambda_hkl_range_from))

        lambda_hkl_range_to = _retrieve_kropff_init_value(
            variable_key=KropffSessionSubKeys.lambda_hkl,
            key=BraggPeakInitParameters.range_to,
        )
        self.parent.ui.lambda_hkl_to_lineEdit.setText(str(lambda_hkl_range_to))

        lambda_hkl_range_step = _retrieve_kropff_init_value(
            variable_key=KropffSessionSubKeys.lambda_hkl,
            key=BraggPeakInitParameters.range_step,
        )
        self.parent.ui.lambda_hkl_step_lineEdit.setText(str(lambda_hkl_range_step))

        # tau
        tau_fix_flag = _retrieve_kropff_init_value(
            variable_key=KropffSessionSubKeys.tau, key=BraggPeakInitParameters.fix_flag
        )
        if tau_fix_flag:
            self.parent.ui.tau_fix_radioButton.setChecked(True)
        else:
            self.parent.ui.tau_range_radioButton.setChecked(True)
        self.parent.kropff_initial_guess_tau_fix_clicked()

        tau_fix_value = _retrieve_kropff_init_value(
            variable_key=KropffSessionSubKeys.tau, key=BraggPeakInitParameters.fix_value
        )
        self.parent.ui.tau_fix_lineEdit.setText(str(tau_fix_value))

        tau_range_from = _retrieve_kropff_init_value(
            variable_key=KropffSessionSubKeys.tau,
            key=BraggPeakInitParameters.range_from,
        )
        self.parent.ui.tau_from_lineEdit.setText(str(tau_range_from))

        tau_range_to = _retrieve_kropff_init_value(
            variable_key=KropffSessionSubKeys.tau, key=BraggPeakInitParameters.range_to
        )
        self.parent.ui.tau_to_lineEdit.setText(str(tau_range_to))

        tau_range_step = _retrieve_kropff_init_value(
            variable_key=KropffSessionSubKeys.tau,
            key=BraggPeakInitParameters.range_step,
        )
        self.parent.ui.tau_step_lineEdit.setText(str(tau_range_step))

        # sigma
        sigma_fix_flag = _retrieve_kropff_init_value(
            variable_key=KropffSessionSubKeys.sigma,
            key=BraggPeakInitParameters.fix_flag,
        )
        if sigma_fix_flag:
            self.parent.ui.sigma_fix_radioButton.setChecked(True)
        else:
            self.parent.ui.sigma_range_radioButton.setChecked(True)
        self.parent.kropff_initial_guess_sigma_fix_clicked()

        sigma_fix_value = _retrieve_kropff_init_value(
            variable_key=KropffSessionSubKeys.sigma,
            key=BraggPeakInitParameters.fix_value,
        )
        self.parent.ui.sigma_fix_lineEdit.setText(str(sigma_fix_value))

        sigma_range_from = _retrieve_kropff_init_value(
            variable_key=KropffSessionSubKeys.sigma,
            key=BraggPeakInitParameters.range_from,
        )
        self.parent.ui.sigma_from_lineEdit.setText(str(sigma_range_from))

        sigma_range_to = _retrieve_kropff_init_value(
            variable_key=KropffSessionSubKeys.sigma,
            key=BraggPeakInitParameters.range_to,
        )
        self.parent.ui.sigma_to_lineEdit.setText(str(sigma_range_to))

        sigma_range_step = _retrieve_kropff_init_value(
            variable_key=KropffSessionSubKeys.sigma,
            key=BraggPeakInitParameters.range_step,
        )
        self.parent.ui.sigma_step_lineEdit.setText(str(sigma_range_step))

        # lambda_hkl = kropff_session_dict[KropffSessionSubKeys.bragg_peak]['lambda_hkl']
        # tau = kropff_session_dict['bragg peak']['tau']
        # sigma = kropff_session_dict['bragg peak']['sigma']
        bragg_peak_tof_graph = kropff_session_dict[KropffTabSelected.bragg_peak]["graph"]

        # self.parent.ui.kropff_bragg_peak_tau_init.setText(tau)
        # index = self.parent.ui.kropff_bragg_peak_sigma_comboBox.findText(sigma)
        # self.parent.ui.kropff_bragg_peak_sigma_comboBox.blockSignals(True)
        # self.parent.ui.kropff_bragg_peak_sigma_comboBox.setCurrentIndex(index)
        # self.parent.ui.kropff_bragg_peak_sigma_comboBox.blockSignals(False)
        if bragg_peak_tof_graph == KropffSessionSubKeys.lambda_hkl:
            self.parent.ui.kropff_lda_hkl_radioButton.setChecked(True)
        elif bragg_peak_tof_graph == KropffSessionSubKeys.tau:
            self.parent.ui.kropff_tau_radioButton.setChecked(True)
        else:
            self.parent.ui.kropff_sigma_radioButton.setChecked(True)

        self.parent.kropff_automatic_threshold_finder_algorithm = kropff_session_dict.get(
            KropffSessionSubKeys.automatic_bragg_peak_threshold_algorithm,
            KropffThresholdFinder.sliding_average,
        )

        icon = QIcon(settings_image)
        self.parent.ui.automatic_bragg_peak_threshold_finder_settings.setIcon(icon)

        self.parent.ui.kropff_fitting_conditions_pushButton.setIcon(icon)
        self.parent.ui.kropff_fitting_conditions_pushButton_1.setIcon(icon)

        self.parent.ui.automatic_bragg_peak_threshold_finder_pushButton.setStyleSheet(interact_me_style)
        self.parent.ui.automatic_kropff_fitting_pushButton.setStyleSheet(interact_me_style)

        self.parent.ui.fitting_kropff_bragg_peak_lambda_hkl_groupBox.setTitle("\u03bb\u2095\u2096\u2097")
        self.parent.ui.fitting_kropff_bragg_peak_tau_groupBox.setTitle("\u03c4")
        self.parent.ui.fitting_kropff_bragg_peak_sigma_groupBox.setTitle("\u03c3")

    def ui(self):
        ui_dict = self.grand_parent.session_dict[DataType.fitting]["ui"]

        # splitters
        try:
            splitter_2_size = ui_dict["splitter_2"]
            self.parent.ui.splitter_2.setSizes(splitter_2_size)

            splitter_size = ui_dict["splitter_4"]
            self.parent.ui.splitter_4.setSizes(splitter_size)

            splitter_3_size = ui_dict["splitter_3"]
            if splitter_3_size:
                self.parent.ui.splitter_3.setSizes(splitter_3_size)
            else:
                self.parent.ui.splitter_3.setSizes([150, 800])

            # kropff_top_horizontal_splitter = ui_dict["kropff_top_horizontal_splitter"]
            self.parent.ui.kropff_top_horizontal_splitter.setSizes(splitter_3_size)

        except TypeError:
            logger.info("Splitters have not been set due to log file format error! This should only show up once.")

    def matplotlib(self):
        def _matplotlib(parent=None, widget=None):
            sc = MplCanvas(parent, width=5, height=2, dpi=100)
            # sc.axes.plot([0,1,2,3,4,5], [10, 1, 20 ,3, 40, 50])
            toolbar = NavigationToolbar(sc, parent)
            layout = QVBoxLayout()
            layout.addWidget(toolbar)
            layout.addWidget(sc)
            widget.setLayout(layout)
            return sc

        self.parent.kropff_high_plot = _matplotlib(parent=self.parent, widget=self.parent.ui.high_widget)
        self.parent.kropff_low_plot = _matplotlib(parent=self.parent, widget=self.parent.ui.low_widget)
        self.parent.kropff_bragg_peak_plot = _matplotlib(parent=self.parent, widget=self.parent.ui.bragg_peak_widget)
